"""velph command line tool / velph-init."""

import pathlib
import shutil
from typing import Literal, Optional

import click

from phelel.velph.cli import cmd_root
from phelel.velph.cli.init.init import run_init
from phelel.velph.cli.utils import (
    DefaultCellChoices,
    PrimitiveCellChoice,
    VelphFilePaths,
    VelphInitParams,
)
from phelel.velph.utils.vasp import VaspPotcar


#
# velph init
#
@cmd_root.command("init")
@click.argument("cell_filename", nargs=1, type=click.Path())
@click.argument("project_folder", nargs=1, type=click.Path())
@click.option(
    "--amplitude",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Distance of displacements in Angstrom. "
        f"(amplitude: float, default={VelphInitParams.amplitude})"
    ),
)
@click.option(
    "--cell-for-nac",
    "cell_for_nac",
    type=str,
    default=None,
    help=(
        'Cell choice for NAC, "primitive" or "unitcell" '
        f'(cell_for_nac: str, default="{DefaultCellChoices.nac.value}")'
    ),
)
@click.option(
    "--cell-for-relax",
    "cell_for_relax",
    type=str,
    default=None,
    help=(
        'Cell choice for relax, "primitive" or "unitcell" '
        f'(cell_for_relax: str, default="{DefaultCellChoices.relax.value}")'
    ),
)
@click.option(
    "--diagonal/--parallel-to-axes",
    "diagonal",
    type=bool,
    default=None,
    help=(
        "Generate displacements only along axes or not."
        f"(diagonal: bool, default={VelphInitParams.diagonal})"
    ),
)
@click.option(
    "--force",
    "force_create",
    is_flag=True,
    default=None,
    help="Create velph.toml even if it already exists.",
)
@click.option(
    "--kspacing",
    "kspacing",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Define k-point grid by KSPACING. "
        f"(kspacing: float, default={VelphInitParams.kspacing})"
    ),
)
@click.option(
    "--kspacing-dense",
    "kspacing_dense",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Define dense k-point grid by KSPACING_DENSE. "
        f"(kspacing_dense: float, default={VelphInitParams.kspacing_dense})"
    ),
)
@click.option(
    "--magmom",
    "magmom",
    type=str,
    default=None,
    help=(
        'String corresponding to INCAR MAGMOM tag value, e.g., "24*1" or "0 0 1"'
        f"(magmom: str, default={VelphInitParams.magmom})"
    ),
)
@click.option(
    "--max-num-atoms",
    "max_num_atoms",
    nargs=1,
    default=None,
    type=int,
    help=(
        "Determine supercell dimension so that number of atoms in supercell "
        "is less than this number. "
        f"(max_num_atoms: int, default={VelphInitParams.max_num_atoms})"
    ),
)
@click.option(
    "--no-find-primitive",
    "find_primitive",
    is_flag=True,
    flag_value=False,
    default=None,
    help=(
        "Disable finding primitive cell in input cell. "
        f"(find_primitive: bool, default={VelphInitParams.find_primitive})"
    ),
)
@click.option(
    "--phelel-dir-name",
    "phelel_dir_name",
    type=str,
    default="phelel",
    help=(
        'Used for backward compatibility, for which set "supercell". '
        '(phelel_dir_name: str, default="phelel")'
    ),
)
@click.option(
    "--phelel-nosym",
    "phelel_nosym",
    is_flag=True,
    default=None,
    help=(
        'Invoke "phelel --nosym". '
        f"(phelel_nosym: bool, default={VelphInitParams.phelel_nosym})"
    ),
)
@click.option(
    "--phonon-max-num-atoms",
    "phonon_max_num_atoms",
    nargs=1,
    default=None,
    type=int,
    help=(
        "Determine phonon supercell dimension so that number of atoms in supercell "
        "for phonon is less than this number if different one from electron-phonon "
        "(phelel) or phonon-phonon (phono3py) is expected. "
        f"(phonon_max_num_atoms: int, default={VelphInitParams.phonon_max_num_atoms})"
    ),
)
@click.option(
    "--plusminus/--auto",
    "plusminus",
    type=bool,
    default=None,
    help=(
        "Plus-minus displacements in supercell, otherwise auto. "
        f"(plusminus: bool, default={VelphInitParams.plusminus})"
    ),
)
@click.option(
    "--primitive-cell-choice",
    "primitive_cell_choice",
    type=str,
    default=None,
    help=(
        'Primitive cell choice, "standardized" or "reduced" '
        "(primitive_cell_choice: str, "
        f'default="{PrimitiveCellChoice.STANDARDIZED.value}")'
    ),
)
@click.option(
    "--symmetrize-cell",
    "symmetrize_cell",
    is_flag=True,
    default=None,
    help=(
        "Symmetrize input crystal structure. "
        f"(symmetrize_cell: bool, default={VelphInitParams.symmetrize_cell})"
    ),
)
@click.option(
    "--template-toml",
    "template_toml_filename",
    nargs=1,
    type=click.Path(),
    default=None,
    help=(
        "File name of template in toml to update velph.toml in python's "
        "dict.update() style."
    ),
)
@click.option(
    "--tolerance",
    "tolerance",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Tolerance to find crystal symmetry. "
        f"(tolerance: float, default={VelphInitParams.tolerance})"
    ),
)
@click.option(
    "--toml-filename",
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default=None,
    help="File name of velph.toml type file to be created.",
)
@click.option(
    "--use-grg",
    "use_grg",
    is_flag=True,
    default=None,
    help=(
        "Use generalized regular grid. "
        f"(use_grg: bool, default={VelphInitParams.use_grg})"
    ),
)
@click.help_option("-h", "--help")
def cmd_init(
    amplitude: Optional[float],
    cell_filename: str,
    cell_for_nac: Optional[Literal["primitive", "unitcell"]],
    cell_for_relax: Optional[Literal["primitive", "unitcell"]],
    find_primitive: Optional[bool],
    force_create: Optional[bool],
    diagonal: Optional[bool],
    plusminus: Optional[bool],
    kspacing: Optional[float],
    kspacing_dense: Optional[float],
    magmom: Optional[str],
    max_num_atoms: Optional[int],
    phelel_dir_name: str,
    phelel_nosym: Optional[bool],
    phonon_max_num_atoms: Optional[int],
    primitive_cell_choice: Optional[str],
    project_folder: str,
    symmetrize_cell: Optional[bool],
    template_toml_filename: Optional[str],
    toml_filename: Optional[str],
    tolerance: Optional[float],
    use_grg: Optional[bool],
):
    """Initialize an electron phonon calculation project.

    Crystal structure (CELL_FILENAME) and new folder where new velph project
    is created have to be specified as command-line arguments.

    Some command options can be specified in the [init.options] section of the
    velph.toml-template file. Each option's key and its corresponding type are
    indicated in parentheses within echo help documentation, for example,

    \b
    [init.options]
    kspacing = 0.2
    kspacing_dense = 0.1
    max_num_atoms = 120

    """  # noqa: D301
    if not pathlib.Path(cell_filename).exists():
        click.echo(f'"{cell_filename}" not found.', err=True)
        return

    vip_cmd_options = {
        "amplitude": amplitude,
        "cell_for_nac": cell_for_nac,
        "cell_for_relax": cell_for_relax,
        "find_primitive": find_primitive,
        "diagonal": diagonal,
        "plusminus": plusminus,
        "kspacing": kspacing,
        "kspacing_dense": kspacing_dense,
        "magmom": magmom,
        "max_num_atoms": max_num_atoms,
        "phelel_nosym": phelel_nosym,
        "phonon_max_num_atoms": phonon_max_num_atoms,
        "primitive_cell_choice": primitive_cell_choice,
        "symmetrize_cell": symmetrize_cell,
        "tolerance": tolerance,
        "use_grg": use_grg,
    }

    cell_filepath = pathlib.Path(cell_filename)
    if cell_filepath.exists():
        vfp_dict = {"cell_filepath": cell_filepath}
    else:
        click.echo(f'"{cell_filename}" not found.', err=True)
        return

    project_folder_path = pathlib.Path(project_folder)
    if project_folder_path.exists():
        if project_folder_path.samefile(pathlib.Path()):
            click.echo(f'Project directory: "{project_folder}".')
        else:
            click.echo(f'File or folder "{project_folder}" already exists.')
            return

    if template_toml_filename is not None:
        velph_tmpl_filepath = pathlib.Path(template_toml_filename)
        if velph_tmpl_filepath.exists():
            vfp_dict["velph_template_filepath"] = velph_tmpl_filepath
        else:
            click.echo(f'"{template_toml_filename}" not found.', err=True)
            return

    if toml_filename is None:
        toml_filepath = project_folder_path / "velph.toml"
    else:
        toml_filepath = project_folder_path / toml_filename

    if toml_filepath.exists():
        if force_create:
            click.echo(f'"{toml_filepath}" exists, but will be overwritten.', err=True)
        else:
            click.echo(
                f'"{toml_filepath}" was not overwritten because it exists.', err=True
            )
            return

    vfp = VelphFilePaths(**vfp_dict)
    toml_lines = run_init(vip_cmd_options, vfp, phelel_dir_name=phelel_dir_name)

    # Write velph.toml.
    if toml_lines:
        if not project_folder_path.exists():
            project_folder_path.mkdir()
            click.echo(f'Created new folder "{project_folder}".')
        with open(toml_filepath, "w") as w:
            w.write("\n".join(toml_lines))
            w.write("\n")
        click.echo(f'Initial settings were written to "{toml_filepath}".')
        if pathlib.Path("POTCAR").exists():
            vasp_potcar = VaspPotcar("POTCAR")
            click.echo('Found "POTCAR".')
            for elem in vasp_potcar.titel:
                print(f"  {elem}")
            enmax = max(vasp_potcar.enmax)
            click.echo(f'  Max ENMAX in "POTCAR" is {enmax}.')
            if project_folder is not None:
                if not (project_folder_path / "POTCAR").exists():
                    shutil.copy2(pathlib.Path("POTCAR"), project_folder_path / "POTCAR")
                    click.echo(f'"POTCAR" was copied to "{project_folder}/POTCAR".')
    else:
        click.echo("")
        click.echo(f'"{toml_filepath}" was not created.')



"""Phelel loader."""

from __future__ import annotations

import os
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import phonopy.cui.load_helper as phonopy_load_helper
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_primitive_matrix

from phelel import Phelel
from phelel.file_IO import read_phelel_params_hdf5
from phelel.interface.phelel_yaml import PhelelYaml
from phelel.interface.vasp.derivatives import read_files


def load(
    phonopy_yaml: Optional[Union[str, bytes, os.PathLike]] = None,
    supercell_matrix: Optional[Union[Sequence, np.ndarray]] = None,
    primitive_matrix: Optional[Union[str, Sequence, np.ndarray]] = None,
    phonon_supercell_matrix: Optional[Union[Sequence, np.ndarray]] = None,
    is_nac: bool = True,
    fft_mesh: Optional[Union[Sequence, np.ndarray]] = None,
    unitcell: Optional[PhonopyAtoms] = None,
    supercell: Optional[PhonopyAtoms] = None,
    dir_names: Optional[Sequence[Union[str, bytes, os.PathLike]]] = None,
    phonon_dir_names: Optional[Sequence[Union[str, bytes, os.PathLike]]] = None,
    unitcell_filename: Optional[Union[str, bytes, os.PathLike]] = None,
    supercell_filename: Optional[Union[str, bytes, os.PathLike]] = None,
    force_sets_filename: Optional[Union[str, bytes, os.PathLike]] = None,
    force_constants_filename: Optional[Union[str, bytes, os.PathLike]] = None,
    subtract_rfs: bool = True,
    symprec: float = 1e-5,
    is_symmetry: bool = True,
    log_level: int = 0,
) -> Phelel:
    """Loader function.

    phonopy_yaml : str, optional
        Filename of "phonopy.yaml"-like file. If this is given, the data
        in the file are parsed. Default is None.
    supercell_matrix : array_like, optional
        Supercell matrix multiplied to input cell basis vectors.
        shape=(3, ) or (3, 3), where the former is considered a diagonal
        matrix. Default is the unit matrix.
        dtype=int
    primitive_matrix : array_like or str, optional
        Primitive matrix multiplied to input cell basis vectors. Default is
        None, which is equivalent to 'auto'.
        For array_like, shape=(3, 3), dtype=float.
        When 'F', 'I', 'A', 'C', or 'R' is given instead of a 3x3 matrix,
        the primitive matrix for the character found at
        https://spglib.github.io/spglib/definition.html
        is used.
    phonon_supercell_matrix : array_like, optional
        Supercell matrix used for phonon calculation. Supercell matrix for
        derivatives of potentials and phonon (derivative of forces) can be
        different. Unless setting this, supercell_matrix is used as
        phonon_supercell_matrix. Default is None.
    is_nac : bool, optional
        If True, look for 'BORN' file. If False, NAS is turned off.
        Default is True.
    fft_mesh : array_like, optional
        FFT mesh numbers for primitive cell used for generating local potential
        derivative interpolation grid. This has to be set to run displacement
        derivative calculations not only dV/du, but also dDij/du. Default is None.
        dtype='int_', shape=(3,)
    unitcell : PhonopyAtoms, optional
        Input unit cell. Default is None.
    supercell : PhonopyAtoms, optional
        Input supercell. With given, default value of primitive_matrix is set
        to 'auto' (can be overwitten). supercell_matrix is ignored. Default is
        None.
    dir_names : list of str, optional
        List of supercell directory names, e.g.,
            ["perfect", "disp-001", "disp-002"]
        Default is None.
    phonon_dir_names : list of str, optional
        List of supercell directory names for phonon calculation, e.g.,
            ["phonon-perfect", "phonon-disp-001", "phonon-disp-002"]
        Default is None.
    unitcell_filename : str, optional
        Input unit cell filename. Default is None.
    supercell_filename : str, optional
        Input supercell filename. When this is specified, supercell_matrix is
        ignored. Default is None.
    subtract_rfs : bool, optional
        Subtract residual forces of perfect supercell from forces of displaced
        supercells. Default is True.
    symprec : float, optional
        Symmetry tolerance used to search crystal symmetry. Default
        is 1e-5.
    is_symmetry : bool, optional
        Use crystal symmetry or not. Default is True.
    log_level : int, optional
        Log level. 0 is most quiet. Default is 0.

    """
    if phonopy_yaml is not None:
        (
            cell,
            smat,
            pmat,
            ph_smat,
            dataset,
            phonon_dataset,
            _,
            _nac_params,
            _,
        ) = _read_phelel_yaml(
            phonopy_yaml, primitive_matrix, None, is_nac, None, symprec
        )
    else:
        cell, smat, pmat = phonopy_load_helper.get_cell_settings(
            supercell_matrix=supercell_matrix,
            primitive_matrix=primitive_matrix,
            unitcell=unitcell,
            supercell=supercell,
            unitcell_filename=unitcell_filename,
            supercell_filename=supercell_filename,
            symprec=symprec,
        )
        ph_smat = phonon_supercell_matrix
        dataset = None
        phonon_dataset = None
        _nac_params = None

    phelel = Phelel(
        cell,
        supercell_matrix=smat,
        primitive_matrix=pmat,
        phonon_supercell_matrix=ph_smat,
        fft_mesh=fft_mesh,
        symprec=symprec,
        is_symmetry=is_symmetry,
        log_level=log_level,
    )
    if dataset:
        phelel.dataset = dataset
    if phonon_dataset:
        phelel.phonon_dataset = phonon_dataset
    elif dataset:
        phelel.phonon_dataset = {k: v for k, v in dataset.items()}
    if _nac_params:
        phelel.nac_params = _nac_params

    try:
        dDVdu, dDijdu, _fft_mesh, fc = read_phelel_params_hdf5(log_level=log_level)
        phelel.dVdu = dDVdu
        phelel.dDijdu = dDijdu
        if fft_mesh is None:
            phelel.fft_mesh = _fft_mesh
        elif not np.array_equal(_fft_mesh, fft_mesh):
            raise RuntimeError("FFT mesh is inconsistent.")
        if fc is not None:
            phelel.force_constants = fc
    except FileNotFoundError:
        pass

    if dir_names:
        phe_input = read_files(
            phelel,
            dir_names,
            phonon_dir_names=phonon_dir_names,
            subtract_rfs=subtract_rfs,
            log_level=log_level,
        )
        if phelel.fft_mesh is not None:
            phelel.run_derivatives(phe_input)
    elif force_constants_filename is not None or force_sets_filename is not None:
        phonopy_load_helper.set_dataset_and_force_constants(
            phelel.phonon,
            None,
            None,
            force_constants_filename=force_constants_filename,
            force_sets_filename=force_sets_filename,
            log_level=log_level,
        )

    return phelel


def _read_phelel_yaml(
    filename, primitive_matrix, nac_params, is_nac, calculator, symprec
):
    """Read properties from phelel.yaml."""
    phe_yml = PhelelYaml().read(filename)
    cell = phe_yml.unitcell
    smat = phe_yml.supercell_matrix
    ph_smat = phe_yml.phonon_supercell_matrix
    if primitive_matrix is not None:
        pmat = get_primitive_matrix(primitive_matrix, symprec=symprec)
    else:
        pmat = phe_yml.primitive_matrix
    if nac_params is not None:
        _nac_params = nac_params
    elif is_nac:
        _nac_params = phe_yml.nac_params
    else:
        _nac_params = None
    dataset = phe_yml.dataset
    ph_dataset = phe_yml.phonon_dataset
    fc = phe_yml.force_constants
    if calculator is None:
        _calculator = phe_yml.calculator
    else:
        _calculator = calculator
    return cell, smat, pmat, ph_smat, dataset, ph_dataset, fc, _nac_params, _calculator



"""Implementation of velph-phono3py-generate."""

from __future__ import annotations

import pathlib

import click
import phono3py
import tomli

from phelel.velph.cli.phelel.generate import (
    write_phonon_supercells,
    write_supercells,
)


def write_supercell_input_files(
    toml_filename: pathlib.Path,
    phono3py_yaml_filename: pathlib.Path,
) -> None:
    """Generate supercells."""
    if not phono3py_yaml_filename.exists():
        click.echo(f'File "{phono3py_yaml_filename}" not found.', err=True)
        click.echo('Run "velph phono3py init" if necessary.', err=True)
        return None

    ph3py = phono3py.load(phono3py_yaml_filename)
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    write_supercells(ph3py, toml_dict, dir_name="phono3py")
    if ph3py.phonon_supercell_matrix is not None:
        if "phonon" in toml_dict["vasp"]["phono3py"]:
            write_phonon_supercells(ph3py, toml_dict, dir_name="phono3py")
        else:
            print(f'[vasp.phono3py.phonon.*] not found in "{toml_filename}"')



"""velph command line tool / velph-selfenergy."""

import pathlib

import click

from phelel.velph.cli import cmd_root
from phelel.velph.cli.selfenergy.generate import write_input_files
from phelel.velph.cli.utils import check_fft


@cmd_root.group("selfenergy")
@click.help_option("-h", "--help")
def cmd_selfenergy():
    """Choose selfenergy options."""
    pass


@cmd_selfenergy.command("generate")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.option(
    "--dry-run/--no-dry-run",
    "-d",
    "dry_run",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str, dry_run: bool):
    """Generate elph input files."""
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    write_input_files(pathlib.Path(toml_filename), dry_run)


@cmd_selfenergy.command("check-fft")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.help_option("-h", "--help")
def cmd_check_fft(toml_filename: str):
    """Show [NGX, NGY, NGZ] in vasprun.xml."""
    check_fft(toml_filename, "selfenergy")



"""Routines to handle spinor rotation."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import numpy as np

from phelel.utils.spherical_harmonics import get_n_and_rotation_order


class SxSySzMatrices:
    """Create matrix representation of <sigma'|Sx|sigma> and so forth.

    hbar is not multiplied.

    The matrix convention is different from that of Pauli matrices.
    m=-1/2 and 1/2 correspond to 0 and 1 indices of arrays here, respectively.

    """

    def __init__(self):
        """Init method."""
        self._Sx = None
        self._Sy = None
        self._Sz = None

    @property
    def Sx(self) -> Optional[np.ndarray]:
        """Return <sigma'|Sx|sigma> for all sigma channel."""
        return self._Sx

    @property
    def Sy(self) -> Optional[np.ndarray]:
        """Return <sigma'|Sy|sigma> for all sigma channel."""
        return self._Sy

    @property
    def Sz(self) -> Optional[np.ndarray]:
        """Return <sigma'|Sz|sigma> for all sigma channel."""
        return self._Sz

    def run(self):
        """Calculate <sigma'|Sx|sigma>, and so on for all sigma channel."""
        self._Sx = self._get_Sx()
        self._Sy = self._get_Sy()
        self._Sz = self._get_Sz()
        return self

    def _get_Sx(self) -> np.ndarray:
        sx = np.zeros((2, 2), dtype=complex, order="C")
        j_s = 1.0 / 2
        for ip, _ in enumerate((-j_s, j_s)):
            for i, m_s in enumerate((-j_s, j_s)):
                if ip == i + 1:
                    sx[ip, i] += np.sqrt((j_s - m_s) * (j_s + m_s + 1))
                if ip == i - 1:
                    sx[ip, i] += np.sqrt((j_s + m_s) * (j_s - m_s + 1))
        return sx / 2

    def _get_Sy(self) -> np.ndarray:
        sy = np.zeros((2, 2), dtype=complex, order="C")
        j_s = 1.0 / 2
        for ip, _ in enumerate((-j_s, j_s)):
            for i, m_s in enumerate((-j_s, j_s)):
                if ip == i + 1:
                    sy[ip, i] -= np.sqrt((j_s - m_s) * (j_s + m_s + 1))
                if ip == i - 1:
                    sy[ip, i] += np.sqrt((j_s + m_s) * (j_s - m_s + 1))
        return 1j * sy / 2

    def _get_Sz(self) -> np.ndarray:
        sz = np.zeros((2, 2), dtype=complex, order="C")
        j_s = 1.0 / 2
        for ip, _ in enumerate((-j_s, j_s)):
            for i, m_s in enumerate((-j_s, j_s)):
                if ip == i:
                    sz[ip, i] = m_s
        return sz


@dataclass(frozen=True)
class PauliMatrices:
    """Pauli matrices.

    hbar is not multiplied.

    Matrix convention is different from SxSySzMatrices. Here, up and down spins
    are assigned to m=1/2 and -1/2, respectively. This means,

        Sgx = [[0, 1], [1, 0]] @ SxSySzMatrices.Sx @ [[0, 1], [1, 0]] * 2
        Sgy = [[0, 1], [1, 0]] @ SxSySzMatrices.Sy @ [[0, 1], [1, 0]] * 2
        Sgz = [[0, 1], [1, 0]] @ SxSySzMatrices.Sz @ [[0, 1], [1, 0]] * 2

    """

    Sgx: tuple = ((0, 1), (1, 0))
    Sgy: tuple = ((0, -1j), (1j, 0))
    Sgz: tuple = ((1, 0), (0, -1))


class SpinorRotationMatrices:
    r"""Rotation matrix for spinor, D\sigma'\sigma(n,alpha).

    D^{\sigma',\sigma(\mathbf{n}, \alpha) =
        <\sigma'|e^{-i/\hbar \mathbf{L}\cdot\mathbf{n}\alpha}|\sigma>.

    By default, the 2x2 matrix is arranged with spin-up at index-0 and spin-down at
    index-1. With `sxsysz`, it is arranged with spin-up at index-1 and spin-down at
    index-0.

    """

    def __init__(
        self,
        R: np.ndarray,
        lattice: np.ndarray,
        sxsysz: Optional[SxSySzMatrices] = None,
    ):
        """Init method.

        Parameters
        ----------
        R : ndarray
            Rotation matrix wrt basis vectors (not Cartesian).
            shape=(3, 3).
        lattice : ndarray.
            Basis vectors in column vectors.
            shape=(3, 3)
        sxsysz : SxSySzMatrices, optional
            Generators of spin rotation. If this is not given, PauliMatrices is used.

        """
        if sxsysz is None:
            self._sxsysz = PauliMatrices()
            self._sx = np.array(self._sxsysz.Sgx) / 2
            self._sy = np.array(self._sxsysz.Sgy) / 2
            self._sz = np.array(self._sxsysz.Sgz) / 2
        else:
            self._sxsysz = sxsysz
            self._sx = self._sxsysz.Sx
            self._sy = self._sxsysz.Sy
            self._sz = self._sxsysz.Sz
        self._d = None
        self._n_vec, self._r_order, self._detR = get_n_and_rotation_order(R, lattice)
        if self._r_order == 1:
            self._alpha = 0.0
        else:
            self._alpha = 2 * np.pi / self._r_order

    @property
    def Delta(self) -> np.ndarray:
        """Return rotation matrix of spinor."""
        return self._Delta

    def run(self):
        """Rotate spinor matrix."""
        self._compute_d()
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        self._Delta = np.array(self._d, dtype=dtype, order="C")
        return self

    def _compute_d(self):
        r"""Calculate <sigma'|e^iS.n\alpha|sigma>."""
        self._d = []
        n = self._n_vec
        sn = n[0] * self._sx + n[1] * self._sy + n[2] * self._sz
        w, U = np.linalg.eigh(sn)
        self._d = (U * np.exp(-1j * w * self._alpha)) @ U.T.conj()



"""Implementation of velph-transport-plot-transport."""

from __future__ import annotations

import click
import h5py
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np


def plot_transport(f_h5py: h5py.File, plot_filename: str, save_plot: bool = False):
    """Plot transport properties.

    Number of "transport_*" is

    ((N(delta) * N(nbands_sum_array) * N(selfen_approx))
      * N(ncarrier_per_cell) * N(ncarrier_den) * N(mu)) * N(temps)

    N(temps) runs in the inner loop.

    sefeln_approx includes
    - scattering_approximation
      (CRTA, SERTA, ERTA_LAMBDA, ERTA_TAU, MRTA_LAMBDA, MRTA_TAU)
    - static_approximation (True or False)

    In the following codes, N(scattering_approximation) * N(temps) are only considered.

    """
    property_names = (
        "e_conductivity",
        "e_conductivity",
        "e_t_conductivity",
        "peltier",
        "seebeck",
    )

    n_transport = len(
        [
            key
            for key in f_h5py["results/electron_phonon/electrons"]
            if "transport_" in key
        ]
    )
    transports = [
        f_h5py[f"results/electron_phonon/electrons/transport_{n + 1}"]
        for n in range(n_transport)
    ]

    for i, transport in enumerate(transports):
        _show(transport, i + 1)

    n_props = len(property_names)
    _, axs = plt.subplots(
        len(transports),
        n_props,
        figsize=(4 * n_props, 4 * len(transports)),
        squeeze=False,
    )
    last_transport_idx = transports[-1]["id_idx"][:]
    for i, transport in enumerate(transports):
        _plot(
            axs[i, :],
            transport,
            property_names,
            i,
            transport["id_idx"][:],
            last_transport_idx,
        )

    plt.tight_layout()
    if save_plot:
        plt.rcParams["pdf.fonttype"] = 42
        plt.savefig(plot_filename)
        click.echo(f'Transport plot was saved in "{plot_filename}".')
    else:
        plt.show()
    plt.close()


def _plot(
    axs: np.ndarray,
    transport: h5py.Group,
    property_names: tuple,
    index: int,
    transport_idx: np.ndarray,
    last_transport_idx: np.ndarray,
):
    """Plot transport properties.

    Second property with the property name "e_conductivity" is resistivity.

    """
    temps = transport["temps"][:]
    dat_filename = f"transport-{index + 1}.dat"
    properties = []

    label_property_names = []
    for prop_name in property_names:
        if prop_name == "e_conductivity" and prop_name in label_property_names:
            label_property_names.append("e_resistivity")
        else:
            label_property_names.append(prop_name)

    for key, label_prop_name in zip(property_names, label_property_names):
        if label_prop_name == "e_resistivity":
            properties.append([3 / np.trace(tensor) for tensor in transport[key][:]])
        else:
            properties.append([np.trace(tensor) / 3 for tensor in transport[key][:]])

    with open(dat_filename, "w") as w:
        print("# temperature", *label_property_names, file=w)
        for temp, props in zip(temps, np.transpose(properties)):
            print(temp, *props, file=w)
        click.echo(f'Transport data {index + 1} was saved in "{dat_filename}".')

    names = [name.decode("utf-8") for name in transport["id_name"][:]]
    labels = []
    click.echo([(name, int(i)) for name, i in zip(names, transport_idx)])
    for i, name in enumerate(names):
        if last_transport_idx[i] > 1:
            if name == "selfen_approx":
                labels.append(transport["scattering_approximation"][()].decode("utf-8"))

    for i, label_property in enumerate(label_property_names):
        if label_property == "e_conductivity":
            axs[i].semilogy(temps, properties[i], ".-")
        else:
            axs[i].plot(temps, properties[i], ".-")
        axs[i].tick_params(axis="both", which="both", direction="in")
        axs[i].tick_params(axis="y", direction="in")
        axs[i].set_xlabel("temperature (K)")
        if labels:
            axs[i].set_ylabel(f"[{index + 1}] {label_property} ({"-".join(labels)})")
        else:
            axs[i].set_ylabel(f"[{index + 1}] {label_property}")
        axs[i].set_xlim(left=0, right=max(temps))
        axs[i].yaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))
        axs[i].ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
        ax = axs[i].secondary_yaxis("right")
        ax.tick_params(axis="y", which="both", direction="in")
        ax.set_yticklabels([])


def _show(transport: h5py.Group, index: int):
    """Show transport properties.

    ['cell_volume', 'dfermi_tol', 'e_conductivity', 'e_t_conductivity', 'emax',
    'emin', 'energy', 'lab', 'mobility', 'mu', 'n', 'n0', 'ne', 'nedos',
    'nelect', 'nh', 'peltier', 'scattering_approximation', 'seebeck', 'static',
    'tau_average', 'temperature', 'transport_function']

    """
    print(f"- parameters:  # {index}")
    for i, temp in enumerate(transport["temps"][:]):
        print(f"  - temperature: {temp}")
        print(
            "    scattering_approximation:",
            transport["scattering_approximation"][()].decode("utf-8"),
        )
        for key in (
            "cell_volume",
            "dfermi_tol",
            ("n0", "number_of_electrons_gausslegendre"),
            "nedos",
        ):
            if isinstance(key, tuple):
                print(f"    {key[1]}: {transport[key[0]][()]}")
            else:
                print(f"    {key}: {transport[key][()]}")
        print(f"    static_approximation: {bool(transport["static"][()])}")

        print("    data_scalar:")
        for key in (
            ("emax", "emax_for_transport_function"),
            ("emin", "emin_for_transport_function"),
            ("mu", "chemical_potential"),
            "n",
            ("ne", "ne_in_conduction_band"),
            ("nh", "nh_in_valence_band"),
            "tau_average",
        ):
            if isinstance(key, tuple):
                print(f"      {key[1]}: {transport[key[0]][i]}")
            else:
                print(f"      {key}: {transport[key][i]}")

        print("    data_array_diagonal:")
        for key in (
            "e_conductivity",
            "e_t_conductivity",
            "mobility",
            "peltier",
            "seebeck",
        ):
            v = transport[key][:].ravel()
            print(
                f"      {key}: [{v[0]:.3e}, {v[4]:.3e}, {v[8]:.3e}, "
                f"{v[5]:.3e}, {v[2]:.3e}, {v[1]:.3e}]"
            )

        print("    data_array_shapes:")
        for key in (
            "energy",
            ("lab", "Onsager_coefficients"),
            "transport_function",
        ):
            if isinstance(key, tuple):
                print(f"      {key[1]}: {list(transport[key[0]][i].shape)}")
            else:
                print(f"      {key}: {list(transport[key][i].shape)}")



"""velph command line tool / velph-relax."""

import pathlib

import click

from phelel.velph.cli import cmd_root
from phelel.velph.cli.relax.generate import write_input_files


@cmd_root.group("relax")
@click.help_option("-h", "--help")
def cmd_relax():
    """Choose relax options."""
    pass


@cmd_relax.command("generate")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str):
    """Generate relax input files."""
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    if not pathlib.Path(toml_filename).exists():
        click.echo(f'"{toml_filename}" not found.', err=True)
        return None

    prev_directory = None
    for i in range(1, 101):
        directory = pathlib.Path(f"relax/iter{i}")
        if directory.exists():
            click.echo(f'"{directory}" exists.')
        else:
            break
        prev_directory = directory

    write_input_files(
        pathlib.Path(toml_filename), directory, prev_directory=prev_directory
    )



"""Phelel API."""

from __future__ import annotations

import io
import os
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional, Union

import numpy as np
from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive, Supercell, isclose
from phonopy.structure.symmetry import Symmetry

from phelel.base.Dij_qij import DDijQij
from phelel.base.local_potential import DLocalPotential
from phelel.file_IO import write_phelel_params_hdf5
from phelel.version import __version__


@dataclass
class PhelelDataset:
    """Data structure of input data to run derivatives."""

    local_potentials: list[np.ndarray]
    Dijs: list[np.ndarray]
    qijs: list[np.ndarray]
    lm_channels: list[dict]
    dataset: Optional[dict] = None
    phonon_dataset: Optional[dict] = None
    forces: Optional[np.ndarray] = None


class Phelel:
    """Phelel class.

    Attributes
    ----------
    unitcell : PhonopyAtoms
        Unit cell.
    supercell : Supercell
        Spercell.
    primitive : Primitive
        Primitive cell.
    symmetry : Symmetry
        Symmetry of supercell.
    phonon : Phonopy
        Phonopy class instance.
    supercell_matrix : ndarray
        Supercell matrix relative to unit cell.
        dtype='int_', shape=(3,3)
    primitive_matrix : ndarray
        Primitive matrix relative to unit cell. Default is None.
        dtype='double', shape=(3,3)
    atom_indices_in_derivatives : ndarray
        Atom indices in supercell used for computing derivatives.
        dtype='int_', shape=(atom_indices_in_derivatives,)
    fft_mesh : ndarray
        FFT mesh numbers for primitive cell used for generating local potential
        derivative interpolation grid.
        dtype='int_', shape=(3,)
    dVdu : DLocalPotential
        A class instance to calculate and store dV/du.
    dDijdu : DDijQij
        A class instance to calculate and store dDij/du and dQij/du.

    """

    def __init__(
        self,
        unitcell: PhonopyAtoms,
        supercell_matrix: Optional[Union[Sequence, np.ndarray]] = None,
        primitive_matrix: Optional[Union[str, Sequence, np.ndarray]] = None,
        phonon_supercell_matrix: Optional[
            Union[int, float, Sequence, np.ndarray]
        ] = None,
        fft_mesh: Optional[Union[Sequence, np.ndarray]] = None,
        symprec: float = 1e-5,
        is_symmetry: bool = True,
        calculator: Optional[str] = None,
        nufft: Optional[str] = None,
        finufft_eps: Optional[float] = None,
        log_level: int = 0,
    ):
        """Init method.

        Parameters
        ----------
        unitcell : PhonopyAtoms
            Unit cell.
        supercell_matrix : array_like, optional
            Supercell matrix multiplied to input cell basis vectors.
            shape=(3, ) or (3, 3), where the former is considered a diagonal
            matrix. Default is the unit matrix.
            dtype=int
        primitive_matrix : array_like or str, optional
            Primitive matrix multiplied to input cell basis vectors. Default is
            None, which is equivalent to 'auto'.
            For array_like, shape=(3, 3), dtype=float.
            When 'F', 'I', 'A', 'C', or 'R' is given instead of a 3x3 matrix,
            the primitive matrix for the character found at
            https://spglib.github.io/spglib/definition.html
            is used.
        phonon_supercell_matrix : array_like, optional
            Supercell matrix used for phonon calculation. Supercell matrix for
            derivatives of potentials and phonon (derivative of forces) can be
            different. Unless setting this, supercell_matrix is used as
            phonon_supercell_matrix. Default is None.
        fft_mesh : array_like, optional
            FFT mesh numbers for primitive cell. Default is None.
            dtype='int_', shape=(3,)
        symprec : float, optional
            Symmetry tolerance used to search crystal symmetry. Default
            is 1e-5.
        is_symmetry : bool, optional
            Use crystal symmetry or not. Default is True.
        calculator :
            A dummy parameter.
        nufft : str or None, optional
            'finufft' only. Default is None, which corresponds to 'finufft'.
        finufft_eps : float or None, optional
            Accuracy of finufft interpolation. Default is None, which
            corresponds to 1e-6.
        log_level : int, optional
            Log level. 0 is most quiet. Default is 0.

        """
        self._unitcell = unitcell
        self._symprec = symprec
        self._is_symmetry = is_symmetry
        self._calculator = calculator
        self._nufft = nufft
        self._finufft_eps = finufft_eps
        self._log_level = log_level

        ph = self._get_phonopy(supercell_matrix, primitive_matrix)
        self._primitive_matrix = ph.primitive_matrix
        self._supercell_matrix = ph.supercell_matrix
        self._primitive = ph.primitive
        self._supercell = ph.supercell
        self._symmetry = ph.symmetry
        self._dataset = ph.dataset
        self._atom_indices_in_derivatives = self._primitive.p2s_map

        if phonon_supercell_matrix is None:
            self._phonon = ph
        else:
            self._phonon = self._get_phonopy(phonon_supercell_matrix, primitive_matrix)
            assert isclose(self._primitive, self._phonon.primitive)

        p2s_mat_float = np.linalg.inv(self.primitive.primitive_matrix)
        self._p2s_matrix = np.rint(p2s_mat_float).astype("int_")
        assert (abs(self._p2s_matrix - p2s_mat_float) < 1e-5).all()

        if phonon_supercell_matrix is None:
            self._phonon_supercell_matrix = None
        else:
            self._phonon_supercell_matrix = self._phonon.supercell_matrix

        self._dVdu = None
        self._dDijdu = None

        if fft_mesh is None:
            self._fft_mesh = None
        else:
            self.fft_mesh = fft_mesh

        self._dDijdu = DDijQij(
            self._supercell,
            symmetry=self._symmetry,
            atom_indices=self._atom_indices_in_derivatives,
            verbose=self._log_level > 0,
        )

    @property
    def version(self) -> str:
        """Return phelel release version number.

        str
            Phelel release version number

        """
        return __version__

    @property
    def calculator(self) -> str:
        """Return calculator interface name.

        str
            Calculator name such as 'vasp', 'qe', etc.

        """
        return self._calculator

    @property
    def unitcell(self) -> PhonopyAtoms:
        """Return unitcell."""
        return self._unitcell

    @property
    def supercell(self) -> Supercell:
        """Return supercell."""
        return self._supercell

    @property
    def primitive(self) -> Primitive:
        """Return primitive cell."""
        return self._primitive

    @property
    def p2s_matrix(self):
        """Return supercell matrix."""
        return self._p2s_matrix

    @property
    def dataset(self) -> dict:
        """Setter and getter of potential dataset."""
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: dict):
        self._dataset = dataset

    @property
    def atom_indices_in_derivatives(self) -> np.ndarray:
        """Return atom indices used for calculation of derivatives."""
        return self._atom_indices_in_derivatives

    @property
    def symmetry(self) -> Symmetry:
        """Return symmetry of supercell."""
        return self._symmetry

    @property
    def supercell_matrix(self) -> np.ndarray:
        """Return supercell matrix."""
        return self._supercell_matrix

    @property
    def primitive_matrix(self) -> Optional[np.ndarray]:
        """Return primitive matrix."""
        return self._primitive_matrix

    @property
    def phonon_supercell_matrix(self) -> Optional[np.ndarray]:
        """Return supercell matrix used for phonon calculation.

        If ``phonon_supercell_matrix`` is not specified at instantiation of
        ``Phelel`` class, this returns None.

        """
        return self._phonon_supercell_matrix

    @property
    def phonon_supercell(self) -> Supercell:
        """Return phonon supercell."""
        return self._phonon.supercell

    @property
    def phonon_primitive(self) -> Primitive:
        """Return phonon primitive cell."""
        return self._phonon.primitive

    @property
    def phonon_dataset(self) -> dict:
        """Setter and getter of phonon dataset."""
        return self._phonon.dataset

    @phonon_dataset.setter
    def phonon_dataset(self, phonon_dataset):
        self._phonon.dataset = phonon_dataset

    @property
    def nac_params(self) -> Optional[dict]:
        """Setter and getter of parameters for non-analytical term correction.

        dict
            Parameters used for non-analytical term correction
            'born': ndarray
                Born effective charges
                shape=(primitive cell atoms, 3, 3), dtype='double', order='C'
            'factor': float
                Unit conversion factor
            'dielectric': ndarray
                Dielectric constant tensor
                shape=(3, 3), dtype='double', order='C'

        """
        return self._phonon._nac_params

    @nac_params.setter
    def nac_params(self, nac_params: dict):
        self._phonon.nac_params = nac_params

    @property
    def force_constants(self) -> Optional[np.ndarray]:
        """Return force constants."""
        return self._phonon.force_constants

    @force_constants.setter
    def force_constants(self, force_constants: np.ndarray):
        self._phonon.force_constants = force_constants

    @property
    def unit_conversion_factor(self) -> float:
        """Return phonon frequency unit conversion factor.

        float
            Phonon frequency unit conversion factor. This factor
            converts sqrt(<force>/<distance>/<AMU>)/2pi/1e12 to THz
            (ordinary frequency).

        """
        return self._phonon.unit_conversion_factor

    @property
    def phonon(self) -> Phonopy:
        """Return Phonopy class instance."""
        return self._phonon

    @property
    def fft_mesh(self) -> np.ndarray:
        """Setter and getter of FFT mesh numbers."""
        return self._fft_mesh

    @fft_mesh.setter
    def fft_mesh(self, fft_mesh: Union[Sequence, np.ndarray]):
        self._fft_mesh = np.array(fft_mesh, dtype="int_")
        self._dVdu = DLocalPotential(
            self._fft_mesh,
            self._p2s_matrix,
            self._supercell,
            symmetry=self._symmetry,
            atom_indices=self._atom_indices_in_derivatives,
            nufft=self._nufft,
            finufft_eps=self._finufft_eps,
            verbose=self._log_level > 0,
        )

    @property
    def dVdu(self) -> Optional[DLocalPotential]:
        """Return DLocalPotential class instance."""
        return self._dVdu

    @dVdu.setter
    def dVdu(self, dVdu: DLocalPotential):
        self._dVdu = dVdu

    @property
    def dDijdu(self) -> DDijQij:
        """Return DDijQij class instance."""
        return self._dDijdu

    @dDijdu.setter
    def dDijdu(self, dDijdu: DDijQij):
        self._dDijdu = dDijdu

    @property
    def supercells_with_displacements(self) -> Optional[list[PhonopyAtoms]]:
        """Return supercells with displacements.

        list of PhonopyAtoms
            Supercells with displacements generated by
            Phonopy.generate_displacements.

        """
        ph = self._get_phonopy(self._supercell_matrix, self._primitive_matrix)
        ph.dataset = self._dataset
        return ph.supercells_with_displacements

    @property
    def phonon_supercells_with_displacements(self) -> Optional[list[PhonopyAtoms]]:
        """Return supercells with displacements for phonons.

        list of PhonopyAtoms
            Supercells with displacements generated by
            Phonopy.generate_displacements.

        """
        return self._phonon.supercells_with_displacements

    def generate_displacements(
        self,
        distance=0.01,
        is_plusminus="auto",
        is_diagonal=True,
    ):
        """Generate displacement dataset."""
        ph = self._get_phonopy(self._supercell_matrix, self._primitive_matrix)
        ph.generate_displacements(
            distance=distance, is_plusminus=is_plusminus, is_diagonal=is_diagonal
        )
        self._dataset = ph.dataset

    def run_derivatives(self, phe_input: PhelelDataset):
        """Run displacement derivatives calculations from temporary raw data.

        Note
        ----
        After calculation, temporary raw data may be deleted.
        Force constants are created to have full matrix shape.

        """
        if self._fft_mesh is None:
            msg = (
                "fft_mesh for dV/du interpolation has to be set before running this "
                "method."
            )
            raise RuntimeError(msg)

        if phe_input.dataset is not None:
            self._dataset = phe_input.dataset
        loc_pots = phe_input.local_potentials
        Dijs = phe_input.Dijs
        qijs = phe_input.qijs

        if phe_input.phonon_dataset is not None:
            self._prepare_phonon(
                dataset=phe_input.phonon_dataset,
                forces=phe_input.forces,
                calculate_full_force_constants=True,
            )
        else:
            self._prepare_phonon(
                dataset=self._dataset,
                forces=phe_input.forces,
                calculate_full_force_constants=True,
            )
        self._dVdu.run(loc_pots[0], loc_pots[1:], self._dataset["first_atoms"])
        self._dDijdu.run(
            Dijs[0],
            Dijs[1:],
            qijs[0],
            qijs[1:],
            self._dataset["first_atoms"],
            phe_input.lm_channels,
        )

    def save_hdf5(
        self, filename: Union[str, bytes, os.PathLike, io.IOBase] = "phelel_params.hdf5"
    ):
        """Write phelel_params.hdf5."""
        write_phelel_params_hdf5(
            dVdu=self._dVdu,
            dDijdu=self._dDijdu,
            supercell_matrix=self._supercell_matrix,
            primitive_matrix=self._primitive_matrix,
            primitive=self._primitive,
            unitcell=self._unitcell,
            supercell=self._supercell,
            atom_indices_in_derivatives=self._atom_indices_in_derivatives,
            disp_dataset=self._dataset,
            force_constants=self._phonon.force_constants,
            phonon_supercell_matrix=self._phonon.supercell_matrix,
            phonon_primitive=self._phonon.primitive,
            phonon_supercell=self._phonon.supercell,
            nac_params=self._phonon.nac_params,
            symmetry_dataset=self._phonon.primitive_symmetry.dataset,
            filename=filename,
        )

    def _prepare_phonon(
        self,
        dataset: Optional[dict] = None,
        forces: Optional[
            Union[
                Sequence,
                np.ndarray,
            ]
        ] = None,
        force_constants: Optional[np.ndarray] = None,
        calculate_full_force_constants: bool = True,
        fc_calculator: Optional[str] = None,
        fc_calculator_options: Optional[str] = None,
        show_drift: bool = True,
    ):
        """Initialize phonon calculation."""
        if dataset is not None:
            self._phonon.dataset = dataset
        if forces is not None:
            self._phonon.forces = forces
            self._phonon.produce_force_constants(
                calculate_full_force_constants=calculate_full_force_constants,
                fc_calculator=fc_calculator,
                fc_calculator_options=fc_calculator_options,
                show_drift=show_drift,
            )
        if force_constants is not None:
            self._phonon.force_constants = force_constants

    def _get_phonopy(
        self,
        supercell_matrix: Optional[Union[int, float, Sequence, np.ndarray]],
        primitive_matrix: Optional[Union[str, Sequence, np.ndarray]],
    ) -> Phonopy:
        """Return Phonopy instance.

        Note
        ----
        store_dense_svecs=False is necessary to be compatible with code in VASP.

        """
        return Phonopy(
            self._unitcell,
            supercell_matrix=supercell_matrix,
            primitive_matrix=primitive_matrix,
            symprec=self._symprec,
            is_symmetry=self._is_symmetry,
            store_dense_svecs=False,
            calculator=self._calculator,
            log_level=self._log_level,
        )



"""velph command line tool / velph-el_bands."""

from __future__ import annotations

import pathlib

import click

from phelel.velph.cli import cmd_root
from phelel.velph.cli.el_bands.generate import write_input_files
from phelel.velph.cli.el_bands.plot import plot_el_bandstructures


@cmd_root.group("el_bands")
@click.help_option("-h", "--help")
def cmd_el_bands():
    """Choose electronic band structure options."""
    pass


#
# velph el_bands generate
#
@cmd_el_bands.command("generate")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str):
    """Generate input files to plot electronic band structure."""
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    write_input_files(pathlib.Path(toml_filename))


#
# velph el_bands plot
#
@cmd_el_bands.command("plot")
@click.option(
    "--window",
    required=True,
    type=(float, float),
    help="Energy window, emin and emax with respect to Fermi level.",
)
@click.help_option("-h", "--help")
def cmd_plot(window: tuple[float, float]):
    """Plot electronic band structure."""
    vaspout_filename_bands = pathlib.Path("el_bands/bands/vaspout.h5")
    vaspout_filename_dos = pathlib.Path("el_bands/dos/vaspout.h5")
    if not vaspout_filename_bands.exists():
        click.echo(f'"{vaspout_filename_bands}" not found.')
    if not vaspout_filename_dos.exists():
        click.echo(f'"{vaspout_filename_dos}" not found.')

    if vaspout_filename_bands.exists() and vaspout_filename_dos.exists():
        plot_el_bandstructures(window, vaspout_filename_bands, vaspout_filename_dos)



"""velph command line tool / velph-phelel."""

from __future__ import annotations

import pathlib
from typing import Optional

import click
import tomli

import phelel
from phelel.cui.phelel_script import finalize_phelel
from phelel.velph.cli import cmd_root
from phelel.velph.cli.phelel.differentiate import run_derivatives
from phelel.velph.cli.phelel.generate import write_supercell_input_files
from phelel.velph.cli.phelel.init import run_init
from phelel.velph.cli.phelel.phonopy import create_phonopy_yaml
from phelel.velph.utils.vasp import CutoffToFFTMesh


@cmd_root.group("phelel")
@click.help_option("-h", "--help")
def cmd_phelel():
    """Choose supercell options."""
    pass


#
# velph phelel init
#
@cmd_phelel.command("init")
@click.argument("toml_filename", nargs=1, type=click.Path(), default="velph.toml")
@click.option(
    "--dir-name",
    "dir_name",
    type=str,
    default="phelel",
    help=(
        'Used for backward compatibility, for which set "supercell". '
        '(dir_name: str, default="phelel")'
    ),
)
@click.help_option("-h", "--help")
def cmd_init(toml_filename: str, dir_name: str):
    """Generate displacements and write phelel_disp.yaml."""
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    phe = run_init(toml_dict)

    phelel_yaml_filename = pathlib.Path(f"{dir_name}/phelel_disp.yaml")
    phelel_yaml_filename.parent.mkdir(parents=True, exist_ok=True)
    finalize_phelel(
        phe,
        displacements_mode=True,
        filename=phelel_yaml_filename,
        sys_exit_after_finalize=False,
    )

    click.echo(f'"{phelel_yaml_filename}" was generated. ')
    click.echo('VASP input files will be generated by "velph phelel generate".')


#
# velph phelel generate
#
@cmd_phelel.command("generate")
@click.argument("toml_filename", nargs=1, type=click.Path(), default="velph.toml")
@click.option(
    "--dir-name",
    "dir_name",
    type=str,
    default="phelel",
    help=(
        'Used for backward compatibility, for which set "supercell". '
        '(phelel_dirname: str, default="phelel")'
    ),
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str, dir_name: str):
    """Generate phelel input files."""
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    yaml_filename = pathlib.Path(f"{dir_name}/phelel_disp.yaml")
    write_supercell_input_files(
        pathlib.Path(toml_filename), yaml_filename, dir_name=dir_name
    )


#
# velph supercell differentiate
#
@cmd_phelel.command("differentiate")
@click.argument("toml_filename", nargs=1, type=click.Path(), default="velph.toml")
@click.option(
    "--dir-name",
    "dir_name",
    type=str,
    default="phelel",
    help=(
        'Used for backward compatibility, for which set "supercell". '
        '(phelel_dirname: str, default="phelel")'
    ),
)
@click.option(
    "--encut",
    nargs=1,
    type=float,
    default=None,
    help=(
        "Cutoff energy corresponding to FFT mesh of local potential grid. "
        "(encut: float, default=None)"
    ),
)
@click.help_option("-h", "--help")
def cmd_differentiate(
    toml_filename: str,
    encut: Optional[float],
    dir_name: str,
) -> None:
    """Calculate derivatives and write phelel_params.hdf5."""
    hdf5_filename = pathlib.Path(f"{dir_name}/phelel_params.hdf5")
    yaml_filename = pathlib.Path(f"{dir_name}/phelel_disp.yaml")

    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    if "phelel" not in toml_dict or "fft_mesh" not in toml_dict["phelel"]:
        click.echo('"fft_mesh" has to be specified in [phelel] section.', err=True)
        click.echo("See velph-hints.", err=True)
        return None

    is_symmetry = True
    try:
        if toml_dict["phelel"]["nosym"] is True:
            click.echo(
                'Found "nosym = true" in [phelel] section. '
                "Symmetrization is turned off."
            )
            is_symmetry = False
    except KeyError:
        pass

    phe = phelel.load(
        yaml_filename,
        fft_mesh=toml_dict["phelel"]["fft_mesh"],
        is_symmetry=is_symmetry,
    )

    if encut is not None:
        try:
            prec = toml_dict["vasp"]["selfenergy"]["incar"]["prec"]
        except KeyError:
            click.echo(f'[vasp.selfenergy.incar] not found in "{toml_filename}".')
            click.echo('prec = "accurate" is assumed.')
            prec = "accurate"
        click.echo(f"FFT mesh is generated for encut={encut}.")
        phe.fft_mesh = CutoffToFFTMesh.get_FFTMesh(encut, phe.primitive.cell, prec)

    if phe.fft_mesh is None:
        click.echo("FFT mesh is not specified.", err=True)
    else:
        if encut is None:
            click.echo(f"FFT mesh: {phe.fft_mesh}.")
        else:
            click.echo(f"FFT mesh: {phe.fft_mesh} (encut={encut}).")

    if run_derivatives(phe, dir_name=dir_name):
        pathlib.Path(hdf5_filename).parent.mkdir(parents=True, exist_ok=True)
        phe.save_hdf5(filename=hdf5_filename)
        click.echo(f'"{hdf5_filename}" has been made.')


#
# velph phelel phonopy
#
@cmd_phelel.command("phonopy")
@click.argument("toml_filename", nargs=1, type=click.Path(), default="velph.toml")
@click.option(
    "--dir-name",
    "dir_name",
    type=str,
    default="phelel",
    help=(
        'Used for backward compatibility, for which set "supercell". '
        '(phelel_dirname: str, default="phelel")'
    ),
)
@click.help_option("-h", "--help")
def cmd_phonopy(toml_filename: str, dir_name: str):
    """Create phonopy_params.yaml."""
    create_phonopy_yaml(
        pathlib.Path(toml_filename),
        pathlib.Path(f"{dir_name}/phelel_disp.yaml"),
        dir_name=dir_name,
    )



"""velph command line tool / velph-hints."""

import click

from phelel.velph.cli import cmd_root


#
# velph hints
#
@cmd_root.command("hints")
def cmd_hints():
    """Show velph command hints."""
    _show_hints("velph.toml")


def _show_hints(toml_filename: str):
    """Show hints."""
    click.echo(
        "------------------------------- velph hints -------------------------------"
    )
    click.echo('To see detailed options, "velph ... --help".')
    click.echo("")
    click.echo("# Initialization")
    click.echo('1. "velph init --help": List velph-init options.')
    click.echo(f'2. "velph init POSCAR_TYPE_FILE FOLDER": Generate "{toml_filename}".')
    click.echo(
        '3. "velph generate": Generate standardized unit cell and primitive cell.'
    )
    click.echo("4. Confirm if space-group-type is the expected one.")
    click.echo('5. Otherwise restart from step 1 using "--tolerance" option.')
    click.echo("")
    click.echo("# Unit cell relaxation (optional)")
    click.echo(f'1. Modify [vasp.relax] section in "{toml_filename}".')
    click.echo('2. "velph relax generate": Generate input files.')
    click.echo("3. Run VASP.")
    click.echo(
        '4. Restart from # Initialization using "CONTCAR" then skip '
        "# Unit cell relaxation."
    )
    click.echo()
    click.echo("# Electronic band structure calculation")
    click.echo(f'1. Modify [vasp.el_bands] sections in "{toml_filename}".')
    click.echo(
        '2. "velph el_bands generate": Generate input files for electronic band '
        "structure."
    )
    click.echo("3. Run VASP in el_bands directories.")
    click.echo(
        '4. "velph el_bands plot --window -3 4" to plot the bands from Ef-3 to Ef+4'
    )
    click.echo()
    click.echo("# NAC calculation (optional)")
    click.echo(f'1. Modify [vasp.nac] section in "{toml_filename}".')
    click.echo('2. "velph nac generate": Generate input files.')
    click.echo("3. Run VASP.")
    click.echo()
    click.echo("# Supercell calculation")
    click.echo(f'1. Modify [vasp.phelel] section in "{toml_filename}".')
    click.echo('2. "velph phelel init": Prepare finite displacement calculation. ')
    click.echo("   NAC parameters are read when NAC calculation is available.")
    click.echo('3. "velph phelel generate": Generate input files.')
    click.echo("4. Run VASP.")
    click.echo()
    click.echo("# Phonon band structure calculation")
    click.echo("Result of supercell calculation is necessary.")
    click.echo(f'1. Modify [vasp.ph_bands] sections in "{toml_filename}".')
    click.echo(
        '2. "velph ph_bands generate": Generate input files for phononic band '
        "structure."
    )
    click.echo("3. Run VASP in ph_bands directories.")
    click.echo('4. "velph ph_bands plot" to plot the phonon bands')
    click.echo()
    click.echo("# Electron transport calculation")
    click.echo(f'1. Write [vasp.transport] section in "{toml_filename}".')
    click.echo('2. (optioanl) "velph transport generate": ' "Dry-run to find FFT-mesh.")
    click.echo("3. (optioanl) Run VASP.")
    click.echo('4. (optioanl) "velph transport check-fft": Check FFT grid.')
    click.echo('5. (optioanl) Modify "fft_mesh" in [phelel] section manually.')
    click.echo('6. "velph supercell differentiate": Generate derivatives hdf5 file')
    click.echo('7. "velph transport generate": Generate input files')
    click.echo()
    click.echo("# Electron self-energy calculation")
    click.echo(f'1. Modify [vasp.selfenergy] section in "{toml_filename}".')
    click.echo(
        '2. (optional) "velph selfenergy generate -d": ' "Dry-run to find FFT-mesh."
    )
    click.echo("3. (optional) Run VASP.")
    click.echo('4. (optioanl) "velph selfenergy check-fft": Check FFT grid.')
    click.echo('5. (optioanl) Modify "fft_mesh" in [phelel] section manually.')
    click.echo('6. "velph phelel differentiate": Generate derivatives hdf5 file')
    click.echo('7. "velph selfenergy generate": Generate input files')
    click.echo()
    click.echo("# Different supercell size for phonon")
    click.echo('1. Write "supercell_dimension" in [phonopy] section and')
    click.echo("2. Write [vasp.phelel.phonon.*] entry.")
    click.echo(
        "------------------------------- velph hints -------------------------------"
    )



"""Phelel file_IO functions."""

from __future__ import annotations

import io
import os
import pathlib
from typing import Optional, Union

import h5py
import numpy as np
from phonopy.structure.atoms import PhonopyAtoms, atom_data
from phonopy.structure.cells import Primitive
from phonopy.structure.symmetry import Symmetry

from phelel.base.Dij_qij import DDijQij
from phelel.base.local_potential import DLocalPotential
from phelel.utils.data import cmplx2real
from phelel.utils.lattice_points import get_lattice_points


def write_phelel_params_hdf5(
    dVdu: Optional["DLocalPotential"] = None,
    dDijdu: Optional["DDijQij"] = None,
    # Rij=None,
    supercell_matrix=None,
    primitive_matrix=None,
    primitive: Optional[Primitive] = None,
    unitcell: Optional[PhonopyAtoms] = None,
    supercell: Optional[PhonopyAtoms] = None,
    atom_indices_in_derivatives=None,
    disp_dataset=None,
    force_constants=None,
    phonon_supercell_matrix=None,
    phonon_primitive: Optional[Primitive] = None,
    phonon_supercell: Optional[PhonopyAtoms] = None,
    nac_params=None,
    symmetry_dataset=None,
    filename="phelel_params.hdf5",
):
    """Write phelel_params.hdf5."""
    with h5py.File(filename, "w") as w:
        _add_datasets(
            w,
            dVdu=dVdu,
            dDijdu=dDijdu,
            supercell_matrix=supercell_matrix,
            primitive_matrix=primitive_matrix,
            primitive=primitive,
            unitcell=unitcell,
            supercell=supercell,
            atom_indices_in_derivatives=atom_indices_in_derivatives,
            disp_dataset=disp_dataset,
            force_constants=force_constants,
            phonon_supercell_matrix=phonon_supercell_matrix,
            phonon_primitive=phonon_primitive,
            phonon_supercell=phonon_supercell,
            nac_params=nac_params,
            symmetry_dataset=symmetry_dataset,
        )


def write_dVdu_hdf5(
    dVdu,
    supercell_matrix,
    primitive_matrix,
    primitive,
    unitcell,
    supercell,
    filename="dVdu.hdf5",
):
    """Write dVdu.hdf5."""
    with h5py.File(filename, "w") as w:
        _add_datasets(
            w,
            dVdu=dVdu,
            supercell_matrix=supercell_matrix,
            primitive_matrix=primitive_matrix,
            primitive=primitive,
            unitcell=unitcell,
            supercell=supercell,
        )


def read_phelel_params_hdf5(
    filename: Union[str, bytes, os.PathLike, io.IOBase] = "phelel_params.hdf5",
    log_level: int = 0,
) -> tuple[DLocalPotential, DDijQij, np.ndarray, np.ndarray]:
    """Read dV/du and dDij/du from phelel_params.hdf5.

    Parameteters
    ------------
    filename :
        File name of phelel_params.hdf5.
    log_level : int
        Log level.

    Returns
    -------
    tuple :
        dDVdu_obj : DLocalPotential
        dDijdu_obj : DDijQij
        fft_mesh : np.ndarray
        fc : np.ndarray, optional

    """
    if pathlib.Path(filename).exists():
        with h5py.File(filename, "r") as f:
            fft_mesh, dVdu, grid_points, lattice_points = read_dVdu_hdf5(f)
            dDijdu, dqijdu, Dij, qij = read_dDijdu_hdf5(f)
            fc = read_force_constants_hdf5(f)
            supercell = PhonopyAtoms(
                cell=f["supercell_lattice"][:].T,
                scaled_positions=f["supercell_positions"][:],
                symbols=[atom_data[n][1] for n in f["supercell_numbers"][:]],
                masses=f["supercell_masses"][:],
            )
            symmetry = Symmetry(supercell)
            if "atom_indices_in_derivatives" in f:
                atom_indices = f["atom_indices_in_derivatives"][:]
            else:
                atom_indices = f["p2s_map"][:]
            if "primitive_matrix" in f:
                pmat = f["primitive_matrix"][:]
            else:
                pmat = (
                    np.linalg.inv(f["supercell_lattice"][:]) @ f["primitive_lattice"][:]
                )
            p2s_mat_float = np.linalg.inv(pmat)
            p2s_matrix = np.rint(p2s_mat_float).astype("int_")
            assert (abs(p2s_matrix - p2s_mat_float) < 1e-5).all()

        if log_level:
            print(f'dV/du was read from "{filename}".')
            print(f'dDij/du was read from "{filename}".')
    else:
        raise FileNotFoundError(f'"{filename}" was not found.')

    dDVdu_obj = DLocalPotential(
        fft_mesh,
        p2s_matrix,
        supercell,
        symmetry=symmetry,
        atom_indices=atom_indices,
    )
    dDVdu_obj.dVdu = dVdu
    dDVdu_obj.grid_points = grid_points
    dDVdu_obj.lattice_points = lattice_points

    dDijdu_obj = DDijQij(
        supercell,
        symmetry=symmetry,
        atom_indices=atom_indices,
    )
    dDijdu_obj.dDijdu = dDijdu
    dDijdu_obj.dqijdu = dqijdu
    dDijdu_obj.Dij = Dij
    dDijdu_obj.qij = qij

    return dDVdu_obj, dDijdu_obj, fft_mesh, fc


def write_dDijdu_hdf5(dDijdu, filename="dDijdu.hdf5"):
    """Write dDijdu.hdf5."""
    with h5py.File(filename, "w") as w:
        _add_datasets(w, dDijdu=dDijdu)


def read_force_constants_hdf5(f):
    """Read force_constants from hdf5 file object."""
    return f["force_constants"][:]


def read_dVdu_hdf5(f):
    """Read dVdu from hdf5 file object."""
    fft_mesh = f["FFT_mesh"][:]
    dVdu = f["dVdu"][:]
    grid_points = f["grid_point"][:]
    lattice_points = f["lattice_point"][:]
    return fft_mesh, dVdu, grid_points, lattice_points


def read_dDijdu_hdf5(f):
    """Read dDijdu from hdf5 file object."""
    dDijdu = f["dDijdu"][:]
    dqijdu = f["dqijdu"][:]
    Dij = f["Dij"][:]
    qij = f["qij"][:]
    return dDijdu, dqijdu, Dij, qij


def _add_datasets(
    w,
    dVdu: Optional["DLocalPotential"] = None,
    dDijdu: Optional["DDijQij"] = None,
    Rij=None,
    supercell_matrix=None,
    primitive_matrix=None,
    primitive: Optional[Primitive] = None,
    unitcell: Optional[PhonopyAtoms] = None,
    supercell: Optional[PhonopyAtoms] = None,
    atom_indices_in_derivatives=None,
    disp_dataset=None,
    force_constants=None,
    phonon_supercell_matrix=None,
    phonon_primitive: Optional[Primitive] = None,
    phonon_supercell: Optional[PhonopyAtoms] = None,
    nac_params=None,
    symmetry_dataset=None,
):
    if dVdu is not None:
        w.create_dataset("dVdu", data=cmplx2real(dVdu.dVdu))
        w.create_dataset("grid_point", data=dVdu.grid_points)
        w.create_dataset("lattice_point", data=dVdu.lattice_points)
        w.create_dataset("FFT_mesh", data=dVdu.fft_mesh)
    if dDijdu is not None:
        w.create_dataset("dDijdu", data=cmplx2real(dDijdu.dDijdu))
        w.create_dataset("dqijdu", data=cmplx2real(dDijdu.dqijdu))
        w.create_dataset("Dij", data=cmplx2real(dDijdu.Dij))
        w.create_dataset("qij", data=cmplx2real(dDijdu.qij))
    if Rij is not None:
        w.create_dataset("Rij", data=cmplx2real(Rij))
    if supercell_matrix is not None:
        w.create_dataset(
            "supercell_matrix", data=np.array(supercell_matrix, dtype="int_", order="C")
        )
    if primitive_matrix is not None:
        w.create_dataset(
            "primitive_matrix",
            data=np.array(primitive_matrix, dtype="double", order="C"),
        )
    if primitive is not None:
        w.create_dataset(
            "primitive_lattice",
            data=np.array(primitive.cell.T, dtype="double", order="C"),
        )
        w.create_dataset(
            "primitive_positions",
            data=np.array(primitive.scaled_positions, dtype="double", order="C"),
        )
        w.create_dataset(
            "primitive_numbers", data=np.array(primitive.numbers, dtype="int_")
        )
        w.create_dataset(
            "primitive_masses", data=np.array(primitive.masses, dtype="double")
        )
        p2s_vectors, p2s_multiplicities = primitive.get_smallest_vectors()
        w.create_dataset("p2s_map", data=np.array(primitive.p2s_map, dtype="int_"))
        w.create_dataset("s2p_map", data=np.array(primitive.s2p_map, dtype="int_"))
        w.create_dataset("shortest_vectors", data=np.array(p2s_vectors, dtype="double"))
        w.create_dataset(
            "shortest_vector_multiplicities",
            data=np.array(p2s_multiplicities, dtype="int_"),
        )
        if atom_indices_in_derivatives is not None:
            if not np.array_equal(atom_indices_in_derivatives, primitive.p2s_map):
                w.create_dataset(
                    "atom_indices_in_derivatives",
                    data=np.array(atom_indices_in_derivatives, dtype="int_"),
                )
    if unitcell is not None:
        w.create_dataset(
            "unitcell_lattice",
            data=np.array(unitcell.cell.T, dtype="double", order="C"),
        )
        w.create_dataset(
            "unitcell_positions",
            data=np.array(unitcell.scaled_positions, dtype="double", order="C"),
        )
        w.create_dataset(
            "unitcell_numbers", data=np.array(unitcell.numbers, dtype="int_")
        )
        w.create_dataset(
            "unitcell_masses", data=np.array(unitcell.masses, dtype="double")
        )
    if supercell is not None:
        w.create_dataset(
            "supercell_lattice",
            data=np.array(supercell.cell.T, dtype="double", order="C"),
        )
        w.create_dataset(
            "supercell_positions",
            data=np.array(supercell.scaled_positions, dtype="double", order="C"),
        )
        w.create_dataset(
            "supercell_numbers", data=np.array(supercell.numbers, dtype="int_")
        )
        w.create_dataset(
            "supercell_masses", data=np.array(supercell.masses, dtype="double")
        )
    if disp_dataset is not None:
        if "first_atoms" in disp_dataset:
            atom_indices = [d["number"] for d in disp_dataset["first_atoms"]]
            w.create_dataset(
                "displacements_atom_indices", data=np.array(atom_indices, dtype="int_")
            )
            disps = [d["displacement"] for d in disp_dataset["first_atoms"]]
            w.create_dataset(
                "displacements_vectors", data=np.array(disps, dtype="double", order="C")
            )
    if force_constants is not None:
        w.create_dataset(
            "force_constants", data=np.array(force_constants, dtype="double", order="C")
        )
    if phonon_supercell_matrix is not None:
        w.create_dataset(
            "phonon_supercell_matrix",
            data=np.array(phonon_supercell_matrix, dtype="int_", order="C"),
        )
    if phonon_primitive is not None:
        p2s_vectors, p2s_multiplicities = phonon_primitive.get_smallest_vectors()
        w.create_dataset(
            "phonon_p2s_map", data=np.array(phonon_primitive.p2s_map, dtype="int_")
        )
        w.create_dataset(
            "phonon_s2p_map", data=np.array(phonon_primitive.s2p_map, dtype="int_")
        )
        w.create_dataset(
            "phonon_shortest_vectors", data=np.array(p2s_vectors, dtype="double")
        )
        w.create_dataset(
            "phonon_shortest_vector_multiplicities",
            data=np.array(p2s_multiplicities, dtype="int_"),
        )
        p2s_mat_float = np.linalg.inv(phonon_primitive.primitive_matrix)
        p2s_matrix = np.rint(p2s_mat_float).astype("int_")
        assert (abs(p2s_matrix - p2s_mat_float) < 1e-5).all()
        lattice_points, _ = get_lattice_points(p2s_matrix)
        w.create_dataset("phonon_lattice_point", data=lattice_points)
    if phonon_supercell is not None:
        w.create_dataset(
            "phonon_supercell_lattice",
            data=np.array(phonon_supercell.cell.T, dtype="double", order="C"),
        )
        w.create_dataset(
            "phonon_supercell_positions",
            data=np.array(phonon_supercell.scaled_positions, dtype="double", order="C"),
        )
        w.create_dataset(
            "phonon_supercell_numbers",
            data=np.array(phonon_supercell.numbers, dtype="int_"),
        )
        w.create_dataset(
            "phonon_supercell_masses",
            data=np.array(phonon_supercell.masses, dtype="double"),
        )
    if nac_params is not None:
        w.create_dataset(
            "born_effective_charges",
            data=np.array(nac_params["born"], dtype="double", order="C"),
        )
        w.create_dataset(
            "dielectric_constant",
            data=np.array(nac_params["dielectric"], dtype="double", order="C"),
        )
    if symmetry_dataset is not None:
        w.create_dataset(
            "transformation_matrix",
            data=np.array(
                symmetry_dataset.transformation_matrix, dtype="double", order="C"
            ),
        )
        w.create_dataset(
            "direct_rotations",
            data=np.array(symmetry_dataset.rotations, dtype="int_", order="C"),
        )
        sym_dataset_dict = vars(symmetry_dataset)
        if "number" in sym_dataset_dict:  # For non-magnetic case
            w.create_dataset(
                "spacegroup_number", data=int(symmetry_dataset.number), dtype="int_"
            )
        elif "uni_number" in sym_dataset_dict:  # For magnetic case
            w.create_dataset(
                "magnetic_spacegroup_uni_number",
                data=int(symmetry_dataset.uni_number),
                dtype="int_",
            )



"""Implementation of velph-transport-generate."""

import pathlib

from phelel.velph.cli.selfenergy.generate import write_selfenergy_input_files


def write_input_files(toml_filepath: pathlib.Path, dry_run: bool):
    """Generate transport input files."""
    write_selfenergy_input_files(toml_filepath, dry_run, calc_type="transport")



"""Generate VASP PROCAR equivalent information."""

import numpy as np


class Projector:
    """Wave function character class."""

    def __init__(self, projectors, lm_orbitals):
        """Init method.

        Parameters
        ----------
        projectors : ndarray
            Wave function charactor <p|Psi>
            dtype=complex128
            shape=(nkpts, nbtot, ispin, nrspinors, nions, lmdim, 4)
            The last index corresponds to usual projector and its derivatives
            along x, y, z.
            Note that exp(ikR) is not multiplied.
        lm_orbitals: dict
            Numbers of orbitals in lm channels of atoms. For example of Si::

               lm_orbitals:
               - atom_index:       1
                 num_l_channels:   4
                 channels:
                 - l :   0
                   m : [   0 ]
                 - l :   0
                   m : [   0 ]
                 - l :   1
                   m : [  -1,   0,   1 ]
                 - l :   1
                   m : [  -1,   0,   1 ]
               - atom_index:       2
                 num_l_channels:   4
                 channels:
                 - l :   0
                   m : [   0 ]
                 - l :   0
                   m : [   0 ]
                 - l :   1
                   m : [  -1,   0,   1 ]
                 - l :   1
                   m : [  -1,   0,   1 ]

        """
        self._proj = projectors
        self._lm_orbitals = lm_orbitals

    def _get_lm_pairs(self, i_atom):
        lm_pairs = []
        lm_to_l = []
        for i, ll in enumerate(self._lm_orbitals[i_atom]["channels"]):
            for m in ll["m"]:
                lm_to_l.append(i)
                lm_pairs.append((ll["l"], m))
        return lm_to_l, lm_pairs

    def _get_channel_pairs(self, lm_pairs):
        channel_pairs = []
        channels = []
        for i, pair_i in enumerate(lm_pairs):
            for j, pair_j in enumerate(lm_pairs):
                if pair_i == pair_j:
                    channel_pairs.append((i, j))
                    channels.append(pair_i)
        return channels, channel_pairs


class CoreCharge:
    """Class to compute core charges.

    Full charge density is given by <phi_k|phi_l> = <~phi_k|S|~phi_l>.
    <~psi_k|S|~psi_l> = <~psi_k|~psi_l> + <~psi_k|p_i> q_ij < p_j|~psi_l>

    This class computes the second term on rhs.

    """

    def __init__(self, projectors, qij):
        """Init method.

        Parameters
        ----------
        projectors : ndarray
            Wave function characters
            dtype=complex128
            shape=(nkpts, nbtot, ispin, nrspinors, nions, lmdim, 4)
        qij : ndarray
            <phi_i|phi_j> - <phi_i~|phi_j~>
            dtype=complex128
            shape=(ncdij, nions, lmdim, lmdim')

        """
        self._proj = projectors
        self._qij = qij

        self._cc = None
        self._run()

    def __str__(self):
        """Return summary of core charge data as text."""
        text = "Core charges\n"
        text += "(nkpts, nbtot) = "
        text += str(self._cc.shape) + "\n"
        for i_k, i_b in np.ndindex(self._cc.shape):
            text += "Sum over left-over charges of cores at "
            text += "k-point %d, band %d: %f\n" % (
                i_k + 1,
                i_b + 1,
                self._cc[i_k, i_b].real,
            )
        return text

    def _run(self):
        proj = self._proj[:, :, :, :, :, :, 0]
        qij = self._qij
        self._cc = np.zeros(proj.shape[0:2], dtype=proj.dtype)
        for i_i in range(proj.shape[4]):
            for i_s in range(proj.shape[2]):
                # Here assume not using non-collinear
                #
                # ncdij = i_s + i_n + j_n*2 (two loops for non-collinear)
                # In fortran
                #   i_s: 1 or 2
                #   i_n: 0 or 1
                #   i_n + j_n * 2: 0, 1, 2, or 3
                #   i_s i_n +j_n * 2: 1, 2, 3, or 4
                for lm, lmp in np.ndindex(qij.shape[2:]):
                    self._cc[:, :] += (
                        proj[:, :, i_s, 0, i_i, lm]
                        * qij[i_s, i_i, lm, lmp]
                        * proj[:, :, i_s, 0, i_i, lmp].conj()
                    )


class Procar(Projector):
    """Class to reproduce VASP PROCAR values."""

    def __init__(self, projectors, lm_orbitals, qtot):
        """Init method.

        Parameters
        ----------
        qtot : ndarray
            <phi_i|phi_j> - <phi~_i|phi~_j>
            dtype=complex128
            shape=(nions, ldim, ldim')

        """
        super(Procar, self).__init__(projectors, lm_orbitals)
        self._qtot = qtot
        self._run()

    def __str__(self):
        """Return array shape of wave function character as text."""
        text = "Wave function character\n"
        text += "(nkpts, nbtot, ispin, nrspinors, nions, lmdim, 4) = "
        text += str(self._proj.shape)
        text += "\n" + "\n".join(self._get_procar_lines())
        return text

    @property
    def procar(self):
        """Return PROCAR value as ndarray.

        The array shape can be seen by __str__.

        """
        return self._procar.real

    def _run(self):
        proj = self._proj[:, :, :, :, :, :, 0]
        procar = np.zeros(
            (proj.shape[2], proj.shape[0], proj.shape[1], proj.shape[4], 9),
            dtype="complex128",
        )

        for i_atom in range(self._proj.shape[4]):
            lm_to_l, lm_pairs = self._get_lm_pairs(i_atom)
            channels, channel_pairs = self._get_channel_pairs(lm_pairs)
            for i_k, i_b, i_s, i_n in np.ndindex(self._proj.shape[0:4]):
                procar[i_s, i_k, i_b, i_atom] += self._get_l_overlap(
                    proj[i_k, i_b, i_s, i_n, i_atom],
                    lm_to_l,
                    channels,
                    channel_pairs,
                    i_atom,
                )
        self._procar = procar

    def _get_procar_lines(self):
        lines = []
        for i_s in range(self._proj.shape[2]):
            for i_k in range(self._proj.shape[0]):
                for i_b in range(self._proj.shape[1]):
                    lines.append("k-point %d, band %d" % (i_k + 1, i_b + 1))
                    for i_atom in range(self._proj.shape[4]):
                        lines.append(
                            self._get_procar_text(
                                i_atom, self._procar[i_s, i_k, i_b, :]
                            )
                        )
                lines.append("")
        return lines

    def _get_procar_text(self, i_atom, procar):
        text = "%3d " % (i_atom + 1)
        text += " ".join(["%6.3f" % x for x in procar[i_atom].real])
        return text

    def _get_l_overlap(self, proj_lm, lm_to_l, channels, channel_pairs, i_atom):
        """Compute <wf-char_m>^* qtot_mm' <wf-char_m'>."""
        p_sum = np.zeros(9, dtype=proj_lm.dtype)
        for lm, c_pair in zip(channels, channel_pairs):
            # print(str(c_pair) + " " + str(lm))
            ll = lm_to_l[c_pair[0]]
            lp = lm_to_l[c_pair[1]]
            qtot = self._qtot[i_atom][ll][lp]
            lm_index = lm[0] ** 2 + lm[1] + lm[0]
            p_sum[lm_index] += proj_lm[c_pair[0]] * qtot * np.conj(proj_lm[c_pair[1]])
        return p_sum


class QTOT:
    """Sum over l channels of <phi_i|phi_j> - <phi~_i|phi~_j>."""

    def __init__(self, qtot):
        """Init method.

        These data are in PAW datasets that are assuemed frozen.

        Parameters
        ----------
        qtot : ndarray
            sum over lm channles of the same l channel of <phi_i|phi_j>
            dtype=complex128
            shape=(nions, ldim, ldim')

        """
        self._qtot = qtot

    def __str__(self):
        """Return q_tot as text."""
        text = "q_tot (nions, ldim, ldim') = "
        text += str(self._qtot.shape) + "\n"
        for i in range(self._qtot.shape[0]):
            text += "Atom %d/%d (%d, %d)\n" % ((i + 1,) + self._qtot.shape)
            for qtot_l in np.transpose(self._qtot[i]):
                text += " ".join(["%10.5f" % x for x in qtot_l])
                text += "\n"
        return text



"""Tests for file_IO functions."""

import io
import pathlib

import h5py
import phonopy
from phonopy.interface.phonopy_yaml import read_cell_yaml

from phelel.file_IO import write_phelel_params_hdf5

cwd_called = pathlib.Path.cwd()


def test_write_phelel_params_hdf5_magnetic_spacegroup_uni_number():
    """Test magnetic_spacegroup_uni_number in write_phelel_params_hdf5."""
    unitcell_str = """unit_cell:
  lattice:
  - [     2.812696943681890,     0.000000000000000,     0.000000000000000 ] # a
  - [     0.000000000000000,     2.812696943681890,     0.000000000000000 ] # b
  - [     0.000000000000000,     0.000000000000000,     2.812696943681890 ] # c
  points:
  - symbol: Cr # 1
    coordinates: [  0.000000000000000,  0.000000000000000,  0.000000000000000 ]
    mass: 51.996100
    reduced_to: 1
    magnetic_moment: 1.00000000
  - symbol: Cr # 2
    coordinates: [  0.500000000000000,  0.500000000000000,  0.500000000000000 ]
    mass: 51.996100
    reduced_to: 1
    magnetic_moment: -1.00000000"""

    cell = read_cell_yaml(io.StringIO(unitcell_str))
    ph = phonopy.Phonopy(cell)
    write_phelel_params_hdf5(symmetry_dataset=ph.primitive_symmetry.dataset)
    for created_filename in ("phelel_params.hdf5",):
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        with h5py.File(file_path, "r") as f:
            assert "magnetic_spacegroup_uni_number" in f
            assert "spacegroup_number" not in f
        file_path.unlink()


def test_write_phelel_params_hdf5_spacegroup_number():
    """Test spacegroup_number in write_phelel_params_hdf5."""
    unitcell_str = """unit_cell:
  lattice:
  - [     2.812696943681890,     0.000000000000000,     0.000000000000000 ] # a
  - [     0.000000000000000,     2.812696943681890,     0.000000000000000 ] # b
  - [     0.000000000000000,     0.000000000000000,     2.812696943681890 ] # c
  points:
  - symbol: Cr # 1
    coordinates: [  0.000000000000000,  0.000000000000000,  0.000000000000000 ]
    mass: 51.996100
    reduced_to: 1
  - symbol: Cr # 2
    coordinates: [  0.500000000000000,  0.500000000000000,  0.500000000000000 ]
    mass: 51.996100
    reduced_to: 1"""

    cell = read_cell_yaml(io.StringIO(unitcell_str))
    ph = phonopy.Phonopy(cell)
    write_phelel_params_hdf5(symmetry_dataset=ph.primitive_symmetry.dataset)
    for created_filename in ("phelel_params.hdf5",):
        file_path = pathlib.Path(cwd_called / created_filename)
        assert file_path.exists()
        with h5py.File(file_path, "r") as f:
            assert "spacegroup_number" in f
            assert "magnetic_spacegroup_uni_number" not in f
        file_path.unlink()



"""velph command line tool / velph-supercell."""

from __future__ import annotations

import pathlib
from typing import Optional

import click
import tomli

from phelel.velph.cli import cmd_root
from phelel.velph.cli.phono3py.generate import write_supercell_input_files
from phelel.velph.cli.phono3py.init import run_init


@cmd_root.group("phono3py")
@click.help_option("-h", "--help")
def cmd_phono3py():
    """Choose phono3py options."""
    pass


#
# velph phono3py init
#
@cmd_phono3py.command("init")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.option(
    "--rd",
    "random_displacements",
    nargs=1,
    default=None,
    type=int,
    help="Number of snapshots of supercells with random directional displacement.",
)
@click.help_option("-h", "--help")
def cmd_init(toml_filename: str, random_displacements: Optional[int]):
    """Generate displacements and write phelel_disp.yaml."""
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    ph3py = run_init(toml_dict, number_of_snapshots=random_displacements)

    phono3py_yaml_filename = pathlib.Path("phono3py/phono3py_disp.yaml")
    phono3py_yaml_filename.parent.mkdir(parents=True, exist_ok=True)
    ph3py.save(phono3py_yaml_filename)

    click.echo(f'"{phono3py_yaml_filename}" was generated. ')
    click.echo('VASP input files will be generated by "velph phono3py generate".')


#
# velph phono3py generate
#
@cmd_phono3py.command("generate")
@click.argument(
    "toml_filename",
    nargs=1,
    type=click.Path(),
    default="velph.toml",
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str):
    """Generate phono3py input files."""
    yaml_filename = "phono3py/phono3py_disp.yaml"
    if not pathlib.Path("POTCAR").exists():
        click.echo('"POTCAR" not found in current directory.')

    write_supercell_input_files(
        pathlib.Path(toml_filename), pathlib.Path(yaml_filename)
    )



"""phelel command line interface."""

from __future__ import annotations

import pathlib
import sys
from typing import Optional, Union

import numpy as np
from phonopy.cui.collect_cell_info import collect_cell_info
from phonopy.cui.phonopy_script import (
    print_end,
    print_error,
    print_error_message,
    print_version,
    store_nac_params,
)
from phonopy.interface.calculator import get_default_physical_units
from phonopy.structure.cells import print_cell

from phelel import Phelel
from phelel.cui.create_supercells import create_phelel_supercells
from phelel.cui.phelel_argparse import get_parser
from phelel.cui.settings import PhelelConfParser
from phelel.interface.phelel_yaml import PhelelYaml
from phelel.version import __version__


# AA is created at http://www.network-science.de/ascii/.
def print_phelel():
    """Show phelel logo."""
    print(
        r"""       _          _      _
 _ __ | |__   ___| | ___| |
| '_ \| '_ \ / _ \ |/ _ \ |
| |_) | | | |  __/ |  __/ |
| .__/|_| |_|\___|_|\___|_|
|_|"""
    )
    print_version(__version__, package_name="phelel", rjust_length=25)


def finalize_phelel(
    phelel: Phelel,
    confs: Optional[dict] = None,
    log_level: Union[int, bool] = 0,
    displacements_mode: bool = False,
    filename: Union[str, pathlib.Path] = "phelel.yaml",
    sys_exit_after_finalize: bool = True,
) -> None:
    """Write phelel.yaml and then exit.

    Parameters
    ----------
    phelel : Phelel
        Phelel instance.
    confs : dict
        This contains the settings and command options that the user set.
    log_level : int
        Log level. 0 means quiet.
    displacements_mode : Bool
        When True, crystal structure is written in the length unit of
        calculator interface in phelel_disp.yaml. Otherwise, the
        default unit (angstrom) is used.
    filename : str, optional
        phelel.yaml is written in this filename.

    """
    if displacements_mode:
        _calculator = phelel.calculator
    else:
        _calculator = None
    _physical_units = get_default_physical_units(_calculator)

    yaml_settings = {"force_sets": False, "displacements": displacements_mode}

    phe_yml = PhelelYaml(
        configuration=confs,
        calculator=_calculator,
        physical_units=_physical_units,
        settings=yaml_settings,
    )
    phe_yml.set_phelel_info(phelel)
    with open(filename, "w") as w:
        w.write(str(phe_yml))

    if log_level > 0:
        print("")
        if displacements_mode:
            print(f'Displacement dataset was written in "{filename}".')
        else:
            print(f'Summary of calculation was written in "{filename}".')
        print_end()

    if sys_exit_after_finalize:
        sys.exit(0)


def main(**argparse_control):
    """Phelel CUI main function."""
    #################
    # Option parser #
    #################
    load_phelel_yaml = argparse_control.get("load_phelel_yaml", False)

    if "args" in argparse_control:  # For pytest
        args = argparse_control["args"]
        log_level = args.log_level
    else:
        parser = get_parser()
        args = parser.parse_args()
        # Set log level
        log_level = 1
        if args.verbose:
            log_level = 2
        if args.quiet:
            log_level = 0
        if args.log_level is not None:
            log_level = args.log_level

    if log_level > 0:
        print_phelel()

    if len(args.filename) > 0:
        phelel_conf = PhelelConfParser(filename=args.filename[0], args=args)
        settings = phelel_conf.settings
    else:
        phelel_conf = PhelelConfParser(args=args)
        settings = phelel_conf.settings

    if settings.symmetry_tolerance is None:
        symprec = 1e-5
    else:
        symprec = settings.symmetry_tolerance

    #####################
    # Initialize phelel #
    #####################
    cell_info = collect_cell_info(
        supercell_matrix=settings.supercell_matrix,
        primitive_matrix=settings.primitive_matrix,
        interface_mode=None,
        cell_filename=settings.cell_filename,
        chemical_symbols=settings.chemical_symbols,
        phonopy_yaml_cls=PhelelYaml,
        load_phonopy_yaml=load_phelel_yaml,
    )
    if "error_message" in cell_info:
        print_error_message(cell_info["error_message"])
        if log_level > 0:
            print_error()
        sys.exit(1)
    unitcell = cell_info["unitcell"]
    supercell_matrix = cell_info["supercell_matrix"]
    primitive_matrix = cell_info["primitive_matrix"]
    unitcell_filename = cell_info["optional_structure_info"][0]
    cell_info["phonon_supercell_matrix"] = settings.phonon_supercell_matrix
    phe_yml = cell_info["phonopy_yaml"]
    if cell_info["phonon_supercell_matrix"] is None and phe_yml:
        ph_smat = phe_yml.phonon_supercell_matrix
        cell_info["phonon_supercell_matrix"] = ph_smat
    phonon_supercell_matrix = cell_info["phonon_supercell_matrix"]

    if settings.create_displacements:
        phelel = create_phelel_supercells(
            cell_info,
            settings,
            symprec,
            load_phelel_yaml=load_phelel_yaml,
            log_level=log_level,
        )
        finalize_phelel(
            phelel,
            confs=phelel_conf.confs,
            log_level=log_level,
            displacements_mode=True,
            filename="phelel_disp.yaml",
        )

    fft_mesh = settings.fft_mesh_numbers
    phelel = Phelel(
        unitcell,
        supercell_matrix,
        primitive_matrix=primitive_matrix,
        phonon_supercell_matrix=phonon_supercell_matrix,
        fft_mesh=fft_mesh,
        symprec=symprec,
        is_symmetry=settings.is_symmetry,
        finufft_eps=settings.finufft_eps,
    )

    if log_level > 0:
        print("")
        print(f'Crystal structure was read from "{unitcell_filename}".')
        print("Settings:")
        if (np.diag(np.diag(supercell_matrix)) - supercell_matrix).any():
            print("  Supercell matrix:")
            for v in supercell_matrix:
                print("    %s" % v)
        else:
            print("  Supercell: %s" % np.diag(supercell_matrix))
        if isinstance(primitive_matrix, str) and primitive_matrix == "auto":
            print("  Primitive matrix (Auto):")
        elif primitive_matrix is not None:
            print("  Primitive matrix:")
        if primitive_matrix is not None:
            for v in phelel.primitive_matrix:
                print("    %s" % v)
        if phonon_supercell_matrix is not None:
            print("  Supercell matrix for phonon:")
            for v in phelel.phonon_supercell_matrix:
                print("    %s" % v)
        if fft_mesh is not None:
            print("  FFT mesh: [%d %d %d]" % tuple(fft_mesh))
        print("Space group type: %s" % phelel.symmetry.get_international_table())
        print("-" * 30 + " primitive cell " + "-" * 30)
        print_cell(phelel.primitive)
        print("-" * 32 + " super cell " + "-" * 32)
        print_cell(phelel.supercell)
        if phelel.phonon_supercell_matrix is not None:
            print("-" * 28 + " phonon super cell " + "-" * 29)
            print_cell(phelel.phonon.supercell)
        print("-" * 76)

    ##################################
    # Create dV/du, dDij/du, dqij/du #
    ##################################
    if True:
        from phelel.interface.vasp.derivatives import create_derivatives

        if phelel.dataset is None or "first_atoms" not in phelel.dataset:
            if cell_info["phonopy_yaml"] is not None:
                phe_yml = cell_info["phonopy_yaml"]
            else:
                if log_level:
                    print(
                        "************************************************************"
                    )
                    print("  Use of POSCAR like crystal structure input is deprected.")
                    print("  Please use phelel-yaml like input.")
                    print(
                        "************************************************************"
                    )
                filename = "phelel_disp.yaml"
                if not pathlib.Path(filename).exists():
                    raise RuntimeError(f'"{filename}" not found.')
                if log_level:
                    print(f'Read displacement datasets from "{filename}".')
                phe_yml = PhelelYaml()
                phe_yml.read(filename)

            phelel.dataset = phe_yml.dataset
            if phe_yml.phonon_dataset is None:
                phelel.phonon_dataset = phe_yml.dataset
            else:
                phelel.phonon_dataset = phe_yml.phonon_dataset

            if pathlib.Path("BORN").exists() or phe_yml.nac_params:
                store_nac_params(
                    phelel.phonon,
                    settings,
                    cell_info["phonopy_yaml"],
                    unitcell_filename,
                    log_level,
                    load_phonopy_yaml=load_phelel_yaml,
                )

        if settings.create_derivatives:
            create_derivatives(
                phelel,
                settings.create_derivatives,
                subtract_rfs=settings.subtract_rfs,
                log_level=log_level,
            )
            if phelel.fft_mesh is not None:
                phelel.save_hdf5(filename="phelel_params.hdf5")
                if log_level > 0:
                    print('"phelel_params.hdf5" has been created.')
            print_end()
            sys.exit(0)

    if log_level > 0:
        finalize_phelel(
            phelel,
            confs=phelel_conf.confs,
            log_level=log_level,
            filename="phelel.yaml",
        )



"""Utilities to handle VASP files."""

from __future__ import annotations

import io
import os
from collections import defaultdict
from typing import Optional, Union

import h5py
import numpy as np
from phonopy.structure.cells import PhonopyAtoms


class VaspKpoints:
    """KPOINTS file writer.

    At least following three formats must be supported.

    ---------------------------------------------
    # Generated by velph
    0
    Gamma
            18            18            18
    0.000000000   0.000000000   0.000000000
    ---------------------------------------------

    ---------------------------------------------
    # Generated by velph
    0
    Reciprocal
    0.000000000   0.045454545   0.045454545
    0.045454545   0.000000000   0.045454545
    0.045454545   0.045454545   0.000000000
    0.000000000   0.000000000   0.000000000
    ---------------------------------------------

    ---------------------------------------------
    k points along high symmetry lines by velph
    51
    line mode
    fractional
    0.00000000 0.00000000 0.00000000 GAMMA
    0.50000000 0.00000000 0.50000000 X

    0.50000000 0.00000000 0.50000000 X
    0.62500000 0.25000000 0.62500000 U

    0.62500000 0.25000000 0.62500000 U
    0.37500000 0.37500000 0.75000000 K

    0.37500000 0.37500000 0.75000000 K
    0.00000000 0.00000000 0.00000000 GAMMA
    ---------------------------------------------

    """

    @classmethod
    def write_line_mode(
        cls,
        fp: Union[str, bytes, os.PathLike, io.IOBase],
        cell: PhonopyAtoms,
        toml_dict: dict,
    ) -> None:
        """Write KPOINTS with line mode."""
        lines = cls._get_line_mode(toml_dict, cell)
        cls._write_lines(fp, lines)

    @staticmethod
    def _write_lines(fp, lines):
        if isinstance(fp, io.IOBase):
            fp.write(("\n".join(lines)).encode("utf-8"))
        else:
            with open(fp, "w") as w:
                print("\n".join(lines), file=w)

    @staticmethod
    def _get_line_mode(toml_dict: dict, cell: PhonopyAtoms) -> list:
        """Return KPOINTS lines.

        toml_dict['line'] is necessary.

        """
        if "line" not in toml_dict:
            raise RuntimeError("line in toml_dict is not found.")
        lines = []
        lines.append("k points along high symmetry lines by velph")
        lines.append(str(toml_dict["line"]) + "")
        lines.append("line mode")
        lines.append("fractional")

        if "path" in toml_dict:
            for p in list(toml_dict["path"]):
                for i in range(2):
                    coord = " ".join([f"{toml_dict[p[i]][j]:.8f}" for j in range(3)])
                    lines.append(f"{coord} {p[i]}")
                lines.append("")
        else:
            try:
                import seekpath
            except ImportError as exc:
                raise ModuleNotFoundError("You need to install seekpath.") from exc

            band_path = seekpath.get_path(cell.totuple())
            for p in list(band_path["path"]):
                for i in range(2):
                    coord = " ".join(
                        [f'{band_path["point_coords"][p[i]][j]:.8f}' for j in range(3)]
                    )
                    lines.append(f"{coord} {p[i]}")
                lines.append("")
        return lines

    @classmethod
    def write_mesh_mode(
        cls,
        fp: Union[str, bytes, os.PathLike, io.IOBase],
        toml_dict: dict,
    ):
        """Write KPOINTS with mesh mode."""
        lines = cls._get_mesh_mode(toml_dict)
        cls._write_lines(fp, lines)

    @staticmethod
    def _get_mesh_mode(toml_dict: dict) -> list:
        if "mesh" not in toml_dict:
            raise RuntimeError('"mesh" not found in dict.')

        mesh = toml_dict["mesh"]
        if "shift" in toml_dict:
            shift = toml_dict["shift"]
        else:
            shift = [0, 0, 0]

        lines = ["# Generated by velph"]
        lines.append("0")
        if isinstance(mesh[0], int):
            lines.append("Gamma")
            lines.append(" ".join([f"{v:13d}" for v in mesh]))
            lines.append(" ".join([f"{v:13.9f}" for v in shift]))
        else:
            lines.append("Reciprocal")
            mesh_inv_t = np.linalg.inv(mesh).T
            mesh_inv_t = np.where(np.abs(mesh_inv_t) < 1e-15, 0, mesh_inv_t)
            for vec in mesh_inv_t:
                lines.append(" ".join([f"{v:13.9f}" for v in vec]))
            lines.append(" ".join([f"{v:13.9f}" for v in shift]))
        return lines


class VaspIncar:
    """Write INCAR file and expand INCAR value.

    Supported data types
    --------------------
    int
    float
    str
    bool

    INCAR writer
    ------------
    Tag name is converted to uppercase.
    True is represented by ".TRUE.".
    False is represented by ".FALSE.".
    Tags are sorted by alphabetical order.

    INCAR value expander
    --------------------
    Asterisk (*) is considered to expand INCAR value.

    """

    specific_tags = []

    @classmethod
    def write(cls, fp: Union[str, bytes, os.PathLike, io.IOBase], toml_dict: dict):
        """Write INCAR."""
        lines = cls._get_incar(toml_dict)
        cls._write_lines(fp, lines)

    @staticmethod
    def expand(line: str) -> list[float]:
        """Expand right hand side of INCAR tag line.

        Parameters
        ----------
        line: str
            "3*1 0 0 0 3*2" -> [1., 1., 1., 0., 0., 0., 2., 2., 2.]

        Returns
        -------
        list[float]
            Expanded INCAR value.

        """
        vals = []
        for chunk in line.split():
            if "*" in chunk:
                n, v = chunk.split("*")
                vals += [float(v)] * int(n)
            else:
                vals.append(float(chunk))
        return vals

    @classmethod
    def _get_incar(cls, toml_dict: dict) -> list:
        sorted_list = sorted(toml_dict.items(), key=lambda pair: pair[0].lower())
        lines = []
        for key, value in sorted_list:
            if value is None:
                continue
            if isinstance(value, bool):
                if value:
                    lines.append(f"{key.upper()} = .TRUE.")
                else:
                    lines.append(f"{key.upper()} = .FALSE.")
            elif isinstance(value, list):
                flattened_list = cls._flatten_list(value)
                if key == "magmom":
                    list_str = cls._compress_magmom_values(flattened_list)
                else:
                    list_str = " ".join([f"{v}" for v in flattened_list])
                lines.append(f"{key.upper()} = {list_str}")
            elif isinstance(value, str):
                lines.append(f"{key.upper()} = {value}")
            elif isinstance(value, (float, int)):
                lines.append(f"{key.upper()} = {value}")
            else:
                raise NotImplementedError(
                    f"Handler for value type {type(value)} not implemented."
                )
        return lines

    @staticmethod
    def _write_lines(fp: Union[str, bytes, os.PathLike, io.IOBase], lines: list[str]):
        if isinstance(fp, io.IOBase):
            fp.write(("\n".join(lines)).encode("utf-8"))
        else:
            with open(fp, "w") as w:
                print("\n".join(lines), file=w)

    @staticmethod
    def _compress_magmom_values(values, tolerance=1e-15):
        uniq_vals = [[values[0]]]
        for i, v in enumerate(values[1:]):
            if abs(v - values[i]) < tolerance:
                uniq_vals[-1].append(v)
            else:
                uniq_vals.append([v])
        return " ".join([f"{len(vals)}*{vals[0]}" for vals in uniq_vals])

    @classmethod
    def _flatten_list(cls, values):
        val_list = []
        for v in values:
            if isinstance(v, list):
                val_list += cls._flatten_list(v)
            else:
                val_list.append(v)
        return val_list


class VaspPotcar:
    """POTCAR reader.

    Attributes
    ----------
    enmax : list[float]
        ENMAX of atoms.

    """

    value_types = {"ENMAX": float, "TITEL": str}

    def __init__(self, fp: Union[str, bytes, os.PathLike, io.IOBase]):
        """Init method."""
        self._property_dict = defaultdict(list)
        self._read(fp)

    @property
    def enmax(self) -> list[float]:
        """Return values of ENMAX in POTCAR as a list."""
        return self._property_dict["ENMAX"]

    @property
    def titel(self) -> list[float]:
        """Return values of TITEL in POTCAR as a list."""
        return self._property_dict["TITEL"]

    def _read(self, fp: Union[str, bytes, os.PathLike, io.IOBase]):
        if isinstance(fp, io.IOBase):
            self._parse(fp)
        else:
            with open(fp) as _fp:
                self._parse(_fp)

    def _parse(self, fp: io.IOBase):
        """Parse lines of POTCAR.

        This method will be more complicated to support more properties.

        """
        for line in fp:
            for key in self.value_types:
                if key in line:
                    if key == "ENMAX":
                        ary = line.split()
                        index = ary.index(key)
                        valstr = ary[index + 2].strip(" ;")
                        value = self.value_types[key](valstr)
                        self._property_dict[key].append(value)
                    elif key == "TITEL":
                        ary = line.split("=")
                        self._property_dict[key].append(ary[1].strip())


class CutoffToFFTMesh:
    """Determin NGX, NGY, NGX."""

    @classmethod
    def get_FFTMesh(
        cls, cutoff_eV: float, lattice: np.ndarray, incar_prec: Optional[str] = None
    ) -> np.ndarray:
        """Return FFT mesh corresponding to (NGX, NGY, NGZ).

        The return value has to be dividable by 2 and factorized by 2, 3, 5, and 7.

        Parameters
        ----------
        lattice : np.ndarray
            Basis vectors in row vectors.
        cutoff_eV : float
            Plane wave cutoff energy in eV.
        incar_prec : str or None
            Default is None, corresponding to "normal".
            cutoff_factor is 4 if 'high', 'accurate', or 'single'. Otherwise 3.
            Note that this is VASP6 definition.

        """
        lengths = np.linalg.norm(lattice, axis=1) / 0.529177249
        cutoff = np.sqrt(cutoff_eV / 13.605826) / (2 * np.pi / lengths)
        fft_mesh = np.rint(cutoff * cls._get_cutoff_factor(incar_prec)).astype(int)
        for i in range(3):
            while not cls._is_factorized_by_2357(fft_mesh[i]):
                fft_mesh[i] += 1
        return fft_mesh

    @staticmethod
    def _get_cutoff_factor(prec: Optional[str]) -> int:
        """Return factor to multiply to cutoff.

        Note this is the VASP6 convention.

        """
        if prec == "singlen" or prec is None:
            return 3
        if prec[0].lower() in ("h", "a", "s"):
            return 4
        else:
            return 3

    @staticmethod
    def _is_factorized_by_2357(n: int) -> bool:
        """Check if n can be dividable by 2 and be factorized only by 2, 3, 5, and 7."""
        if (n // 2) * 2 != n:
            return False

        _n = n
        for div in (2, 3, 5, 7):
            while (_n // div) * div == _n:
                _n = _n // div

        if _n == 1:
            return True
        else:
            return False


def read_magmom(magmom: str) -> Optional[list[float]]:
    """Read text file to obtain MAGMOM information.

    Parameters
    ----------
    magmom : str
        String corresponding to INCAR MAGMOM tag value, e.g., "24*1" or "0 0 1".

    """
    return VaspIncar().expand(magmom.strip())


def read_crystal_structure_from_h5(f_vaspout_h5: h5py.File, group: str) -> PhonopyAtoms:
    """Read crystal structure from vaspout.h5."""
    direct = int(f_vaspout_h5[f"{group}/direct_coordinates"][()])
    scale = f_vaspout_h5[f"{group}/scale"][()]
    lattice = f_vaspout_h5[f"{group}/lattice_vectors"][:] * scale
    positions = f_vaspout_h5[f"{group}/position_ions"][:]
    number_ion_types = f_vaspout_h5[f"{group}/number_ion_types"][:]
    ion_types = f_vaspout_h5[f"{group}/ion_types"][:]

    symbols = []
    for symbol, number in zip(ion_types, number_ion_types):
        symbols += [symbol.decode()] * number

    if not direct:
        positions = positions @ np.linalg.inv(lattice)

    assert len(symbols) == len(positions)

    cell = PhonopyAtoms(cell=lattice, scaled_positions=positions, symbols=symbols)

    return cell



"""velph command line tool / velph-generate."""

import pathlib

import click
import tomli
from phonopy.interface.calculator import write_crystal_structure
from phonopy.structure.atoms import parse_cell_dict

from phelel.velph.cli import cmd_root


#
# velph generate
#
@cmd_root.command("generate")
@click.option(
    "-f",
    "toml_filename",
    nargs=1,
    type=click.Path(exists=True),
    default="velph.toml",
    show_default=True,
    help="Specify velph.toml",
)
@click.option(
    "--prefix",
    nargs=1,
    type=click.Path(),
    default="POSCAR",
    show_default=True,
    help="{prefix}-unitcell, {prefix}-primitive",
)
@click.help_option("-h", "--help")
def cmd_generate(toml_filename: str, prefix: str):
    """Write POSCAR-unitcell and POSCAR-primitive.

    Filename prefix "POSCAR" can be replaced using the "prefix" option.

    """
    _run_generate(toml_filename, prefix)


def _run_generate(toml_filename: str, prefix: str) -> None:
    """Generate {prefix}-unitcell and {prefix}-primitive."""
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    filename = f"{prefix}-unitcell"
    _write_cell(filename, toml_dict["unitcell"])
    if "primitive_cell" in toml_dict:
        filename = f"{prefix}-primitive"
        _write_cell(filename, toml_dict["primitive_cell"])


def _write_cell(filename, toml_cell_dict):
    if pathlib.Path(filename).exists():
        click.echo(f'"{filename}" was not overwritten because it exists.', err=True)
    else:
        write_crystal_structure(filename, parse_cell_dict(toml_cell_dict))
        click.echo(f'"{filename}" was generated.', err=True)
